@@ -2,7 +2,7 @@
  *
  * Pigment media rendering library
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -133,7 +133,7 @@
 #include "pgmmarshal.h"
 #include "pgmenumtypes.h"
 
-GST_DEBUG_CATEGORY (pgm_canvas_debug);
+GST_DEBUG_CATEGORY_STATIC (pgm_canvas_debug);
 #define GST_CAT_DEFAULT pgm_canvas_debug
 
 /* Canvas signals */
@@ -189,14 +189,14 @@ pgm_canvas_dispose (GObject *object)
 {
   PgmCanvas *canvas = PGM_CANVAS (object);
 
-  while (canvas->far)
-    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->far->data));
+  while (canvas->far_layer)
+    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->far_layer->data));
 
-  while (canvas->middle)
-    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->middle->data));
+  while (canvas->middle_layer)
+    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->middle_layer->data));
 
-  while (canvas->near)
-    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->near->data));
+  while (canvas->near_layer)
+    pgm_canvas_remove (canvas, PGM_DRAWABLE (canvas->near_layer->data));
 
   g_hash_table_unref (canvas->format_counter);
   canvas->format_counter = NULL;
@@ -289,9 +289,9 @@ pgm_canvas_init (PgmCanvas *canvas)
   guint i;
 
   /* Public data */
-  canvas->far = NULL;
-  canvas->middle = NULL;
-  canvas->near = NULL;
+  canvas->far_layer = NULL;
+  canvas->middle_layer = NULL;
+  canvas->near_layer = NULL;
   canvas->width = 4.0f;
   canvas->height = 3.0f;
   canvas->pixel_formats = 0;
@@ -437,19 +437,20 @@ pgm_canvas_add (PgmCanvas *canvas,
   switch (layer)
     {
     case PGM_DRAWABLE_FAR:
-      canvas->far = g_list_insert_sorted (canvas->far, drawable,
-                                          (GCompareFunc) z_compare);
-      order = g_list_index (canvas->far, drawable);
+      canvas->far_layer = g_list_insert_sorted (canvas->far_layer, drawable,
+                                                (GCompareFunc) z_compare);
+      order = g_list_index (canvas->far_layer, drawable);
       break;
     case PGM_DRAWABLE_MIDDLE:
-      canvas->middle = g_list_insert_sorted (canvas->middle, drawable,
-                                             (GCompareFunc) z_compare);
-      order = g_list_index (canvas->middle, drawable);
+      canvas->middle_layer = g_list_insert_sorted (canvas->middle_layer,
+                                                   drawable,
+                                                   (GCompareFunc) z_compare);
+      order = g_list_index (canvas->middle_layer, drawable);
       break;
     case PGM_DRAWABLE_NEAR:
-      canvas->near = g_list_insert_sorted (canvas->near, drawable,
-                                           (GCompareFunc) z_compare);
-      order = g_list_index (canvas->near, drawable);
+      canvas->near_layer = g_list_insert_sorted (canvas->near_layer, drawable,
+                                                 (GCompareFunc) z_compare);
+      order = g_list_index (canvas->near_layer, drawable);
       break;
     default:
       GST_WARNING_OBJECT (drawable, "has not been assigned to any layer");
@@ -502,6 +503,13 @@ pgm_canvas_remove (PgmCanvas *canvas,
 
   GST_OBJECT_LOCK (drawable);
 
+  /* Check if the drawable is in the canvas */
+  if (G_UNLIKELY (GST_OBJECT_PARENT (drawable) != GST_OBJECT (canvas)))
+    {
+      GST_OBJECT_UNLOCK (drawable);
+      return PGM_ERROR_X;
+    }
+
   /* Check if the drawable is already being removed and immediately
    * return */
   if (G_UNLIKELY (GST_OBJECT_FLAG_IS_SET
@@ -521,13 +529,13 @@ pgm_canvas_remove (PgmCanvas *canvas,
   switch (drawable->layer)
     {
     case PGM_DRAWABLE_FAR:
-      canvas->far = g_list_remove (canvas->far, drawable);
+      canvas->far_layer = g_list_remove (canvas->far_layer, drawable);
       break;
     case PGM_DRAWABLE_MIDDLE:
-      canvas->middle = g_list_remove (canvas->middle, drawable);
+      canvas->middle_layer = g_list_remove (canvas->middle_layer, drawable);
       break;
     case PGM_DRAWABLE_NEAR:
-      canvas->near = g_list_remove (canvas->near, drawable);
+      canvas->near_layer = g_list_remove (canvas->near_layer, drawable);
       break;
     default:
       /* We should have a really big problem to take that path */
@@ -671,15 +679,15 @@ pgm_canvas_set_order (PgmCanvas *canvas,
   switch (drawable->layer)
     {
     case PGM_DRAWABLE_NEAR:
-      list = canvas->near;
+      list = canvas->near_layer;
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      list = canvas->middle;
+      list = canvas->middle_layer;
       break;
 
     case PGM_DRAWABLE_FAR:
-      list = canvas->far;
+      list = canvas->far_layer;
       break;
 
     default:
@@ -719,11 +727,11 @@ pgm_canvas_set_order (PgmCanvas *canvas,
 
   /* Affect the reordered list to the canvas */
   if (drawable->layer == PGM_DRAWABLE_NEAR)
-    canvas->near = list;
+    canvas->near_layer = list;
   else if (drawable->layer == PGM_DRAWABLE_MIDDLE)
-    canvas->middle = list;
+    canvas->middle_layer = list;
   else
-    canvas->far = list;
+    canvas->far_layer = list;
 
   GST_OBJECT_UNLOCK (drawable);
   GST_OBJECT_UNLOCK (canvas);
@@ -779,15 +787,15 @@ pgm_canvas_get_order (PgmCanvas *canvas,
   switch (*layer)
     {
     case PGM_DRAWABLE_NEAR:
-      *order = g_list_index (canvas->near, drawable);
+      *order = g_list_index (canvas->near_layer, drawable);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      *order = g_list_index (canvas->middle, drawable);
+      *order = g_list_index (canvas->middle_layer, drawable);
       break;
 
     case PGM_DRAWABLE_FAR:
-      *order = g_list_index (canvas->far, drawable);
+      *order = g_list_index (canvas->far_layer, drawable);
       break;
 
     default:
@@ -826,15 +834,15 @@ pgm_canvas_get_layer_count (PgmCanvas *canvas,
   switch (layer)
     {
     case PGM_DRAWABLE_NEAR:
-      *count = g_list_length (canvas->near);
+      *count = g_list_length (canvas->near_layer);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      *count = g_list_length (canvas->middle);
+      *count = g_list_length (canvas->middle_layer);
       break;
 
     case PGM_DRAWABLE_FAR:
-      *count = g_list_length (canvas->far);
+      *count = g_list_length (canvas->far_layer);
       break;
 
     default:
@@ -925,30 +933,31 @@ _pgm_canvas_update_order (PgmCanvas *canvas,
   switch (drawable->layer)
     {
     case PGM_DRAWABLE_NEAR:
-      if (!need_ordering_update (canvas->near, drawable))
+      if (!need_ordering_update (canvas->near_layer, drawable))
         goto no_update_needed;
-      canvas->near = g_list_remove (canvas->near, drawable);
-      canvas->near = g_list_insert_sorted (canvas->near, drawable,
-                                           (GCompareFunc) z_compare);
-      order = g_list_index (canvas->near, drawable);
+      canvas->near_layer = g_list_remove (canvas->near_layer, drawable);
+      canvas->near_layer = g_list_insert_sorted (canvas->near_layer, drawable,
+                                                 (GCompareFunc) z_compare);
+      order = g_list_index (canvas->near_layer, drawable);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      if (!need_ordering_update (canvas->middle, drawable))
+      if (!need_ordering_update (canvas->middle_layer, drawable))
         goto no_update_needed;
-      canvas->middle = g_list_remove (canvas->middle, drawable);
-      canvas->middle = g_list_insert_sorted (canvas->middle, drawable,
-                                             (GCompareFunc) z_compare);
-      order = g_list_index (canvas->middle, drawable);
+      canvas->middle_layer = g_list_remove (canvas->middle_layer, drawable);
+      canvas->middle_layer = g_list_insert_sorted (canvas->middle_layer,
+                                                   drawable,
+                                                   (GCompareFunc) z_compare);
+      order = g_list_index (canvas->middle_layer, drawable);
       break;
 
     case PGM_DRAWABLE_FAR:
-      if (!need_ordering_update (canvas->far, drawable))
+      if (!need_ordering_update (canvas->far_layer, drawable))
         goto no_update_needed;
-      canvas->far = g_list_remove (canvas->far, drawable);
-      canvas->far = g_list_insert_sorted (canvas->far, drawable,
-                                          (GCompareFunc) z_compare);
-      order = g_list_index (canvas->far, drawable);
+      canvas->far_layer = g_list_remove (canvas->far_layer, drawable);
+      canvas->far_layer = g_list_insert_sorted (canvas->far_layer, drawable,
+                                                (GCompareFunc) z_compare);
+      order = g_list_index (canvas->far_layer, drawable);
       break;
 
     default:
