@@ -52,15 +52,20 @@ static const GEnumValue _pgm_drawable_property_values[] = {
   { PGM_DRAWABLE_BG_COLOR, "PGM_DRAWABLE_BG_COLOR", "drawable-bg-color" },
   { PGM_DRAWABLE_FG_COLOR, "PGM_DRAWABLE_FG_COLOR", "drawable-fg-color" },
   { PGM_DRAWABLE_OPACITY, "PGM_DRAWABLE_OPACITY", "drawable-opacity" },
+  { PGM_DRAWABLE_REGENERATE, "PGM_DRAWABLE_REGENERATE", "drawable-regenerate" },
   { PGM_IMAGE_DATA_BUFFER, "PGM_IMAGE_DATA_BUFFER", "image-data-buffer" },
   { PGM_IMAGE_DATA_GST_BUFFER, "PGM_IMAGE_DATA_GST_BUFFER", "image-data-gst-buffer" },
-  { PGM_IMAGE_DATA_FD, "PGM_IMAGE_DATA_FD", "image-data-fd" },
+  { PGM_IMAGE_DATA_PIXBUF, "PGM_IMAGE_DATA_PIXBUF", "image-data-pixbuf" },
   { PGM_IMAGE_DATA_IMAGE, "PGM_IMAGE_DATA_IMAGE", "image-data-image" },
   { PGM_IMAGE_DATA_EMPTY, "PGM_IMAGE_DATA_EMPTY", "image-data-empty" },
+  { PGM_IMAGE_MAPPING_MATRIX, "PGM_IMAGE_MAPPING_MATRIX", "image-mapping-matrix" },
   { PGM_IMAGE_ALIGNMENT, "PGM_IMAGE_ALIGNMENT", "image-alignment" },
   { PGM_IMAGE_LAYOUT, "PGM_IMAGE_LAYOUT", "image-layout" },
   { PGM_IMAGE_INTERP, "PGM_IMAGE_INTERP", "image-interp" },
   { PGM_IMAGE_ASPECT_RATIO, "PGM_IMAGE_ASPECT_RATIO", "image-aspect-ratio" },
+  { PGM_IMAGE_BORDER_WIDTH, "PGM_IMAGE_BORDER_WIDTH", "image-border-width" },
+  { PGM_IMAGE_BORDER_INNER_COLOR, "PGM_IMAGE_BORDER_INNER_COLOR", "image-border-inner-color" },
+  { PGM_IMAGE_BORDER_OUTER_COLOR, "PGM_IMAGE_BORDER_OUTER_COLOR", "image-border-outer-color" },
   { PGM_TEXT_LABEL, "PGM_TEXT_LABEL", "text-label" },
   { PGM_TEXT_MARKUP, "PGM_TEXT_MARKUP", "text-markup" },
   { PGM_TEXT_FONT_FAMILY, "PGM_TEXT_FONT_FAMILY", "text-font-family" },
@@ -119,6 +124,7 @@ static const GFlagsValue _pgm_drawable_picking_signal_values[] = {
   { PGM_DRAWABLE_PICKING_DRAG_END, "PGM_DRAWABLE_PICKING_DRAG_END", "drag-end" },
   { PGM_DRAWABLE_PICKING_ENTERED, "PGM_DRAWABLE_PICKING_ENTERED", "entered" },
   { PGM_DRAWABLE_PICKING_LEFT, "PGM_DRAWABLE_PICKING_LEFT", "left" },
+  { PGM_DRAWABLE_PICKING_SCROLLED, "PGM_DRAWABLE_PICKING_SCROLLED", "scrolled" },
   { PGM_DRAWABLE_PICKING_MASK, "PGM_DRAWABLE_PICKING_MASK", "mask" },
   { 0, NULL, NULL }
 };
@@ -148,6 +154,9 @@ static const GEnumValue _pgm_event_type_values[] = {
   { PGM_KEY_RELEASE, "PGM_KEY_RELEASE", "key-release" },
   { PGM_EXPOSE, "PGM_EXPOSE", "expose" },
   { PGM_CONFIGURE, "PGM_CONFIGURE", "configure" },
+  { PGM_DRAG_MOTION, "PGM_DRAG_MOTION", "drag-motion" },
+  { PGM_DRAG_DROP, "PGM_DRAG_DROP", "drag-drop" },
+  { PGM_DRAG_LEAVE, "PGM_DRAG_LEAVE", "drag-leave" },
   { PGM_SCROLL, "PGM_SCROLL", "scroll" },
   { PGM_DELETE, "PGM_DELETE", "delete" },
   { 0, NULL, NULL }
@@ -312,7 +321,7 @@ static const GEnumValue _pgm_image_storage_type_values[] = {
   { PGM_IMAGE_EMPTY, "PGM_IMAGE_EMPTY", "empty" },
   { PGM_IMAGE_BUFFER, "PGM_IMAGE_BUFFER", "buffer" },
   { PGM_IMAGE_GST_BUFFER, "PGM_IMAGE_GST_BUFFER", "gst-buffer" },
-  { PGM_IMAGE_FD, "PGM_IMAGE_FD", "fd" },
+  { PGM_IMAGE_PIXBUF, "PGM_IMAGE_PIXBUF", "pixbuf" },
   { PGM_IMAGE_IMAGE, "PGM_IMAGE_IMAGE", "image" },
   { 0, NULL, NULL }
 };
@@ -499,7 +508,9 @@ pgm_viewport_cursor_get_type (void)
 }
 
 static const GFlagsValue _pgm_viewport_capacity_values[] = {
-  { PGM_VIEWPORT_HARDWARE_ACCELERATION, "PGM_VIEWPORT_HARDWARE_ACCELERATION", "acceleration" },
+  { PGM_VIEWPORT_HARDWARE_ACCELERATION, "PGM_VIEWPORT_HARDWARE_ACCELERATION", "hardware-acceleration" },
+  { PGM_VIEWPORT_APPLICATION_EMBEDDING, "PGM_VIEWPORT_APPLICATION_EMBEDDING", "application-embedding" },
+  { PGM_VIEWPORT_OPACITY, "PGM_VIEWPORT_OPACITY", "opacity" },
   { 0, NULL, NULL }
 };
 
@@ -515,5 +526,27 @@ pgm_viewport_capacity_get_type (void)
 }
 
 
+/* enumerations from "./pgmlinearalgebra.h" */
+#include "./pgmlinearalgebra.h"
+static const GEnumValue _pgm_mat4x4_predefined_values[] = {
+  { PGM_MAT4X4_FLIP_HORIZONTAL, "PGM_MAT4X4_FLIP_HORIZONTAL", "flip-horizontal" },
+  { PGM_MAT4X4_FLIP_VERTICAL, "PGM_MAT4X4_FLIP_VERTICAL", "flip-vertical" },
+  { PGM_MAT4X4_ROTATE_CLOCKWISE, "PGM_MAT4X4_ROTATE_CLOCKWISE", "rotate-clockwise" },
+  { PGM_MAT4X4_ROTATE_COUNTER_CLOCKWISE, "PGM_MAT4X4_ROTATE_COUNTER_CLOCKWISE", "rotate-counter-clockwise" },
+  { 0, NULL, NULL }
+};
+
+GType
+pgm_mat4x4_predefined_get_type (void)
+{
+  static GType type = 0;
+
+  if (!type)
+    type = g_enum_register_static ("PgmMat4x4Predefined", _pgm_mat4x4_predefined_values);
+
+  return type;
+}
+
+
 /* Generated data ends here */
 
