@@ -84,6 +84,10 @@ GType pgm_viewport_cursor_get_type (void);
 #define PGM_TYPE_VIEWPORT_CAPACITY pgm_viewport_capacity_get_type()
 GType pgm_viewport_capacity_get_type (void);
 
+/* --- ./pgmlinearalgebra.h --- */
+#define PGM_TYPE_MAT4X4_PREDEFINED pgm_mat4x4_predefined_get_type()
+GType pgm_mat4x4_predefined_get_type (void);
+
 G_END_DECLS
 
 #endif
