@@ -2,7 +2,7 @@
  *
  * Pigment media rendering library
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -918,8 +918,30 @@ pgm_event_copy (PgmEvent *event)
 void
 pgm_event_free (PgmEvent *event)
 {
-  if (G_LIKELY (event))
-    g_slice_free (PgmEvent, event);
+  if (!G_LIKELY (event))
+    return;
+
+  switch (event->type)
+    {
+      /* The array of strings need to freed in the case of a Dnd event */
+    case PGM_DRAG_MOTION:
+    case PGM_DRAG_DROP:
+    case PGM_DRAG_LEAVE:
+      {
+        PgmEventDnd *dnd = (PgmEventDnd*) event;
+
+        if (dnd->uri)
+          {
+            g_strfreev (dnd->uri);
+            dnd->uri = NULL;
+          }
+      }
+
+    default:
+      break;
+    }
+
+  g_slice_free (PgmEvent, event);
 }
 
 GType
