@@ -2,7 +2,7 @@
  *
  * Pigment GStreamer image sink
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -30,10 +30,17 @@
  * @see_also: #PgmImage
  *
  * #PgmImageSink is a GStreamer image sink element allowing to build GStreamer
- * pipelines outputing frames directly to a #PgmImage object defined as a
- * property.
+ * pipelines outputing frames directly to #PgmImage.
  *
- * Last reviewed on 2007-06-11 (0.1.5)
+ * You can explicitly use the pgm_image_sink_new() function to instanciate a
+ * new sink, but you can alos use the standard GStreamer functions using
+ * gst_element_factory_make() or gst_parse_launch(), the factory name of the
+ * sink being "pgmimagesink".
+ *
+ * Internally, the sink is simply outputing GStreamer buffers (#GstBuffer) to
+ * images (#PgmImage) using the pgm_image_set_from_gst_buffer() function.
+ *
+ * Last reviewed on 2008-01-28 (0.3.4)
  */
 
 #if HAVE_CONFIG_H
@@ -43,6 +50,7 @@
 #include <gst/video/video.h> /* Helper functions */
 #include <string.h>
 #include "pgmimagesink.h"
+#include "pgmcanvas.h"
 
 static GstElementDetails pgm_image_sink_details =
     GST_ELEMENT_DETAILS ("Video Sink", "Sink/Video",
@@ -61,8 +69,8 @@ static GstStaticPadTemplate pgm_image_sink_template =
             "width = (int) [ 1, MAX ], "
             "height = (int) [ 1, MAX ]"));
 
-GST_DEBUG_CATEGORY_STATIC (gst_debug_pgm_image_sink);
-#define GST_CAT_DEFAULT gst_debug_pgm_image_sink
+GST_DEBUG_CATEGORY_STATIC (pgm_image_sink_debug);
+#define GST_CAT_DEFAULT pgm_image_sink_debug
 
 enum {
   ARG_0,
@@ -613,6 +621,9 @@ pgm_image_sink_class_init (PgmImageSinkClass *klass)
   gstbasesink_class->get_times = GST_DEBUG_FUNCPTR (pgm_image_sink_get_times);
   gstbasesink_class->preroll = GST_DEBUG_FUNCPTR (pgm_image_sink_show);
   gstbasesink_class->render = GST_DEBUG_FUNCPTR (pgm_image_sink_show);
+
+  GST_DEBUG_CATEGORY_INIT (pgm_image_sink_debug, "pgm_image_sink", 0,
+                           "pigment image sink gstreamer element");
 }
 
 static void
@@ -648,22 +659,75 @@ pgm_image_sink_get_type (void)
   return sink_type;
 }
 
-/* GStreamer plugin initialization */
+/* PgmImageSink methods */
 
-static gboolean
-plugin_init (GstPlugin *plugin)
+/**
+ * pgm_image_sink_new:
+ * @name: the name of the new image sink.
+ *
+ * Creates a new image sink with the given name.
+ *
+ * Returns: a new #PgmImageSink.
+ */
+GstElement*
+pgm_image_sink_new (const gchar *name)
 {
-  if (!gst_element_register (plugin, "pgmimagesink", GST_RANK_NONE,
-                             PGM_TYPE_IMAGE_SINK))
-    return FALSE;
-
-  GST_DEBUG_CATEGORY_INIT (gst_debug_pgm_image_sink, "pgmimagesink", 0,
-                           "pigment image sink element");
-
-  return TRUE;
+  return gst_element_factory_make ("pgmimagesink", name);
 }
 
-GST_PLUGIN_DEFINE (GST_VERSION_MAJOR, GST_VERSION_MINOR,
-                   "pgmimagesink", "Outputs video to PgmImage object",
-                   plugin_init, VERSION, "LGPL", PACKAGE,
-                   "https://code.fluendo.com/pigment/trac")
+/**
+ * pgm_image_sink_set_image:
+ * @sink: a #PgmImageSink object.
+ * @image: the #PgmImage object.
+ *
+ * Sets @image as the #PgmImage object which will receive the frames
+ * (#GstBuffer) outputed by @sink.
+ *
+ * MT Safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_sink_set_image (PgmImageSink *sink,
+                          PgmImage *image)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE_SINK (sink), PGM_ERROR_X);
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+
+  //GST_OBJECT_LOCK (sink);
+
+  g_object_set (G_OBJECT (sink), "image", image, NULL);
+
+  //GST_OBJECT_UNLOCK (sink);
+
+  return PGM_ERROR_OK;
+}
+
+/**
+ * pgm_image_sink_get_image:
+ * @sink: a #PgmImageSink object.
+ * @image: a pointer to a #PgmImage pointer where the image will be stored.
+ * Unref after usage.
+ *
+ * Retrieves the #PgmImage object in @image which @sink is using to output its
+ * frames (#GstBuffer).
+ *
+ * MT Safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_sink_get_image (PgmImageSink *sink,
+                          PgmImage **image)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE_SINK (sink), PGM_ERROR_X);
+  g_return_val_if_fail (image != NULL, PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (sink);
+
+  *image = gst_object_ref (sink->image);
+
+  GST_OBJECT_UNLOCK (sink);
+
+  return PGM_ERROR_OK;
+}
