@@ -2,7 +2,7 @@
  *
  * Pigment media rendering library
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -51,16 +51,6 @@ G_BEGIN_DECLS
   (G_TYPE_INSTANCE_GET_CLASS((obj), PGM_TYPE_VIEWPORT, PgmViewportClass))
 #define PGM_VIEWPORT_CAST(obj) ((PgmViewport *) (obj))
 
-/**
- * PGM_VIEWPORT_HAS_CAPS:
- * @viewport: a #PgmViewport object.
- * @mask: the #PgmViewportCapacity to check, can be a mask of capacities.
- *
- * Retrieves whether @viewport supports the @mask of #PgmViewportCapacity.
- */
-#define PGM_VIEWPORT_HAS_CAPS(viewport,mask) \
-  ((PGM_VIEWPORT(viewport)->caps_mask) & (mask) == (mask))
-
 /**
  * PgmViewportCursor:
  * @PGM_VIEWPORT_LEFT_ARROW: Standard left arrow system cursor.
@@ -79,13 +69,30 @@ typedef enum {
  * PgmViewportCapacity:
  * @PGM_VIEWPORT_HARDWARE_ACCELERATION: Supports accelerated rendering thanks
  * to dedicated hardware.
+ * @PGM_VIEWPORT_APPLICATION_EMBEDDING: Supports embedding into another
+ * application.
+ * @PGM_VIEWPORT_OPACITY: Supports opacity so that compositing managers can
+ * blend the viewport with other applications.
  *
  * Capacities supported by the viewport.
  */
 typedef enum {
   PGM_VIEWPORT_HARDWARE_ACCELERATION = (1 << 0),
+  PGM_VIEWPORT_APPLICATION_EMBEDDING = (1 << 1),
+  PGM_VIEWPORT_OPACITY               = (1 << 2),
 } PgmViewportCapacity;
 
+/* FIXME: Should not be exported in this header since it's only used
+ *        internally. That could be fixed separating public and private
+ *        GObject fields (see #183) */
+typedef struct __PgmIOWatch _PgmIOWatch;
+struct __PgmIOWatch {
+  gint        fd[2];
+  GIOChannel *out;
+  GIOChannel *in;
+  gint        tag;
+};
+
 /**
  * PgmViewport:
  * @factory: the factory used to instantiate the #PgmViewport.
@@ -94,6 +101,9 @@ typedef enum {
  * @cursor: the current cursor.
  * @fullscreen: the fullscreen state.
  * @visible: the visible state.
+ * @decorated: the decorated state.
+ * @opacity: the opacity.
+ * @alpha_blending: the alpha blending state.
  * @icon: the current icon.
  * @width: the viewport width.
  * @height: the viewport height.
@@ -125,13 +135,16 @@ struct _PgmViewport {
   GdkPixbuf         *icon;
   gboolean           fullscreen;
   gboolean           visible;
+  gboolean           decorated;
+  guchar             opacity;
+  gboolean           alpha_blending;
 
   gint width, height;
   gint width_mm, height_mm;
 
-  PgmMat4x4 projection, inv_projection;
-  gint projected_x, projected_y;
-  gint projected_width, projected_height;
+  PgmMat4x4 *projection, *inv_projection;
+  gint       projected_x, projected_y;
+  gint       projected_width, projected_height;
 
   gfloat pixel_aspect_ratio;
   gfloat viewport_ratio;
@@ -140,15 +153,17 @@ struct _PgmViewport {
   gulong caps_mask;
 
   /*< private >*/
-  /* Event dedicated lock */
-  GMutex *event_lock;
-
   /* Event handling */
-  GList      *event;
-  gint        event_fd[2];
-  GIOChannel *channel_out;
-  GIOChannel *channel_in;
-  gint        source;
+  _PgmIOWatch *event_watch;
+  gboolean     event_watch_up;
+  GMutex      *event_lock;
+  GList       *event_list;
+
+  /* Pixel handling */
+  _PgmIOWatch *pixel_watch;
+  gboolean     pixel_watch_up;
+  GMutex      *pixel_lock;
+  GList       *pixel_list;
 
   /* Signal handler id */
   gulong canvas_size_handler;
@@ -169,14 +184,20 @@ struct _PgmViewport {
  * @scroll_event: the "scroll-event" signal.
  * @motion_notify_event: the "motion-notify-event" signal.
  * @expose_event: the "expose-event" signal.
- * @update: the update virtual method.
+ * @drag_motion_event: the "drag-motion-event" signal.
+ * @drag_drop_event: the "drag-drop-event" signal.
+ * @drag_leave_event: the "drag-leave-event" signal.
+ * @pixels_read: the "pixels-read" signal.
  * @set_title: the set_title virtual method.
  * @show: the show virtual method.
  * @hide: the hide virtual method.
+ * @set_decorated: the set_decorated virtual method.
  * @set_cursor: the set_cursor virtual method.
  * @set_icon: the set_icon virtual method.
  * @set_size: the set_size virtual method.
  * @get_size: the get_size virtual method.
+ * @set_alpha_blending: the set_alpha_blending virtual method.
+ * @set_opacity: the set_opacity virtual method.
  * @set_fullscreen: the set_fullscreen virtual method.
  * @get_fullscreen: the get_fullscreen virtual method.
  * @set_screen_resolution: the set_screen_resolution virtual method.
@@ -185,8 +206,12 @@ struct _PgmViewport {
  * @get_screen_size_mm: the get_screen_size_mm virtual method.
  * @set_canvas: the set_canvas virtual method.
  * @update_projection: the update_projection virtual method.
+ * @get_embedding_id: the get_embedding_id virtual method.
  * @get_pixel_formats: the get_pixel_formats virtual method.
  * @get_caps_mask: the get_caps_mask virtual method.
+ * @get_frame_rate: the get_frame_rate virtual method.
+ * @read_pixels: the read_pixels virtual method.
+ * @set_drag_status: the set_drag_status virtual method.
  *
  * Pigment drawable class.
  */
@@ -197,44 +222,58 @@ struct _PgmViewportClass {
 
   /* signals */
 
-  void (*button_press_event)        (PgmViewport *viewport,
+  void     (*button_press_event)   (PgmViewport *viewport,
                                      PgmEventButton *event);
 
-  void (*button_release_event)      (PgmViewport *viewport,
-                                     PgmEventButton *event);
+  void     (*button_release_event) (PgmViewport *viewport,
+                                    PgmEventButton *event);
 
-  void (*key_press_event)           (PgmViewport *viewport,
-                                     PgmEventKey *event);
+  void     (*key_press_event)      (PgmViewport *viewport,
+                                    PgmEventKey *event);
 
-  void (*key_release_event)         (PgmViewport *viewport,
-                                     PgmEventKey *event);
+  void     (*key_release_event)    (PgmViewport *viewport,
+                                    PgmEventKey *event);
 
-  void (*configure_event)           (PgmViewport *viewport,
-                                     PgmEventConfigure *event);
+  void     (*configure_event)      (PgmViewport *viewport,
+                                    PgmEventConfigure *event);
 
-  void (*delete_event)              (PgmViewport *viewport,
-                                     PgmEventDelete *event);
+  void     (*delete_event)         (PgmViewport *viewport,
+                                    PgmEventDelete *event);
 
-  void (*scroll_event)              (PgmViewport *viewport,
-                                     PgmEventScroll *event);
+  void     (*scroll_event)         (PgmViewport *viewport,
+                                    PgmEventScroll *event);
 
-  void (*motion_notify_event)       (PgmViewport *viewport,
-                                     PgmEventMotion *event);
+  void     (*motion_notify_event)  (PgmViewport *viewport,
+                                    PgmEventMotion *event);
 
-  void (*expose_event)              (PgmViewport *viewport,
-                                     PgmEventExpose *event);
+  void     (*expose_event)         (PgmViewport *viewport,
+                                    PgmEventExpose *event);
+
+  gboolean (*drag_motion_event)    (PgmViewport *viewport,
+                                    PgmEventDnd *event);
+
+  void     (*drag_drop_event)      (PgmViewport *viewport,
+                                    PgmEventDnd *event);
+
+  void     (*drag_leave_event)     (PgmViewport *viewport,
+                                    PgmEventDnd *event);
+
+  void     (*pixels_read)          (PgmViewport *viewport,
+                                    guint width,
+                                    guint height,
+                                    gpointer pixels);
 
   /* virtual methods for subclasses */
 
-  PgmError (*update)                (PgmViewport *viewport);
-
   PgmError (*set_title)             (PgmViewport *viewport,
                                      const gchar *title);
 
   PgmError (*show)                  (PgmViewport *viewport);
-
   PgmError (*hide)                  (PgmViewport *viewport);
 
+  PgmError (*set_decorated)         (PgmViewport *viewport,
+                                     gboolean decorated);
+
   PgmError (*set_cursor)            (PgmViewport *viewport,
                                      PgmViewportCursor cursor);
 
@@ -248,6 +287,12 @@ struct _PgmViewportClass {
                                      gint *width,
                                      gint *height);
 
+  PgmError (*set_alpha_blending)    (PgmViewport *viewport,
+                                     gboolean alpha_blending);
+
+  PgmError (*set_opacity)           (PgmViewport *viewport,
+                                     guchar opacity);
+
   PgmError (*set_fullscreen)        (PgmViewport *viewport,
                                      gboolean fullscreen);
   PgmError (*get_fullscreen)        (PgmViewport *viewport,
@@ -272,36 +317,50 @@ struct _PgmViewportClass {
 
   PgmError (*update_projection)     (PgmViewport *viewport);
 
+  PgmError (*get_embedding_id)      (PgmViewport *viewport,
+                                     gulong *embedding_id);
+
   PgmError (*get_pixel_formats)     (PgmViewport *viewport,
                                      gulong *formats_mask);
 
   PgmError (*get_caps_mask)         (PgmViewport *viewport,
                                      gulong *caps_mask);
 
-  /*< private >*/
-  /* Default projection matrix */
-  PgmMat4x4 default_projection;
+  PgmError (*get_frame_rate)        (PgmViewport *viewport,
+                                     guint *frame_rate);
 
+  PgmError (*read_pixels)           (PgmViewport *viewport,
+                                     guint x,
+                                     guint y,
+                                     guint width,
+                                     guint height,
+                                     guint8 *pixels);
+
+  PgmError (*set_drag_status)       (PgmViewport *viewport,
+                                     gboolean accept);
+
+  /*< private >*/
   /* Padding */
   gpointer _pgm_reserved[PGM_PADDING];
 };
 
 GType        pgm_viewport_get_type              (void);
 
-PgmError     pgm_viewport_update                (PgmViewport *viewport);
-
 PgmError     pgm_viewport_set_title             (PgmViewport *viewport,
                                                  const gchar *title);
 PgmError     pgm_viewport_get_title             (PgmViewport *viewport,
                                                  gchar **title);
 
 PgmError     pgm_viewport_show                  (PgmViewport *viewport);
-
 PgmError     pgm_viewport_hide                  (PgmViewport *viewport);
-
 PgmError     pgm_viewport_is_visible            (PgmViewport *viewport,
                                                  gboolean *visible);
 
+PgmError     pgm_viewport_set_decorated         (PgmViewport *viewport,
+                                                 gboolean decorated);
+PgmError     pgm_viewport_get_decorated         (PgmViewport *viewport,
+                                                 gboolean *decorated);
+
 PgmError     pgm_viewport_set_cursor            (PgmViewport *viewport,
                                                  PgmViewportCursor cursor);
 PgmError     pgm_viewport_get_cursor            (PgmViewport *viewport,
@@ -319,6 +378,16 @@ PgmError     pgm_viewport_get_size              (PgmViewport *viewport,
                                                  gint *width,
                                                  gint *height);
 
+PgmError     pgm_viewport_set_alpha_blending    (PgmViewport *viewport,
+                                                 gboolean alpha_blending);
+PgmError     pgm_viewport_get_alpha_blending    (PgmViewport *viewport,
+                                                 gboolean *alpha_blending);
+
+PgmError     pgm_viewport_set_opacity           (PgmViewport *viewport,
+                                                 guchar opacity);
+PgmError     pgm_viewport_get_opacity           (PgmViewport *viewport,
+                                                 guchar *opacity);
+
 PgmError     pgm_viewport_set_fullscreen        (PgmViewport *viewport,
                                                  gboolean fullscreen);
 PgmError     pgm_viewport_get_fullscreen        (PgmViewport *viewport,
@@ -364,12 +433,30 @@ PgmError     pgm_viewport_from_canvas           (PgmViewport *viewport,
                                                  gfloat viewport_y,
                                                  gfloat viewport_z);
 
+PgmError     pgm_viewport_get_embedding_id      (PgmViewport *viewport,
+                                                 gulong *embedding_id);
+
 PgmError     pgm_viewport_get_pixel_formats     (PgmViewport *viewport,
                                                  gulong *formats_mask);
 
 PgmError     pgm_viewport_get_caps_mask         (PgmViewport *viewport,
                                                  gulong *caps_mask);
 
+PgmError     pgm_viewport_get_frame_rate        (PgmViewport *viewport,
+                                                 guint *frame_rate);
+
+PgmError     pgm_viewport_read_pixels           (PgmViewport *viewport,
+                                                 guint x,
+                                                 guint y,
+                                                 guint width,
+                                                 guint height,
+                                                 guint8 *pixels);
+
+PgmError     pgm_viewport_push_pixels           (PgmViewport *viewport,
+                                                 guint width,
+                                                 guint height,
+                                                 guint8 *pixels);
+
 G_END_DECLS
 
 #endif /* __PGM_VIEWPORT_H__ */
