@@ -2,7 +2,7 @@
  *
  * Pigment OpenGL plugin
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -33,8 +33,8 @@
 #include <Carbon/Carbon.h>
 #include "pgmaglbackend.h"
 
-GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
-#define GST_CAT_DEFAULT pgm_gl_debug
+GST_DEBUG_CATEGORY_STATIC (pgm_gl_aglbackend_debug);
+#define GST_CAT_DEFAULT pgm_gl_aglbackend_debug
 
 static void set_foreground();
 static void init_agl(PgmAglBackend *aglbackend);
@@ -156,7 +156,7 @@ pgm_agl_backend_dispatch_events (PgmBackend *backend)
   EventRef theEvent;
   EventTargetRef theTarget;
 
-  GST_DEBUG_OBJECT (backend, "dispatch_events");
+  GST_LOG_OBJECT (backend, "dispatch_events");
   CHECK_MAIN_EVENT_LOOP();
 
   theTarget = GetEventDispatcherTarget();
@@ -164,13 +164,13 @@ pgm_agl_backend_dispatch_events (PgmBackend *backend)
   while ((err = ReceiveNextEvent(0, NULL, kEventDurationNoWait, true,
                                  &theEvent)) == noErr)
   {
-    GST_DEBUG_OBJECT(backend, "!!! DISPATCHING");
+    GST_LOG_OBJECT(backend, "!!! DISPATCHING");
     SendEventToEventTarget(theEvent, theTarget);
     ReleaseEvent(theEvent);
   }
 
   if (err != eventLoopTimedOutErr)
-    GST_DEBUG_OBJECT(backend, "ReceiveNextEvent = %ld", err);
+    GST_LOG_OBJECT(backend, "ReceiveNextEvent = %ld", err);
 
   return TRUE;
 }
@@ -207,8 +207,7 @@ app_event_handler(EventHandlerCallRef handler,
   UInt32 kind = GetEventKind(event);
   HICommand command;
 
-  GST_WARNING_OBJECT(user_data, "app_event_handler");
-  GST_WARNING("app_event_handler");
+  GST_LOG_OBJECT(user_data, "app_event_handler");
 
   if (window) {
     GetWindowPortBounds(window, &rectPort);
@@ -333,25 +332,25 @@ window_event_handler(EventHandlerCallRef handler,
 
   PgmEvent *pgmevent = NULL;
 
-  GST_DEBUG_OBJECT(NULL, "window_event_handler");
+  GST_LOG("window_event_handler");
 
   switch (class) {
   case kEventClassKeyboard:
     switch (kind) {
     case kEventRawKeyDown:
-      GST_WARNING_OBJECT(NULL, "window_event_handler :: kEventRawKeyDown");
+      GST_LOG("window_event_handler :: kEventRawKeyDown");
       pgmevent = pgm_event_new (PGM_KEY_PRESS);
       translate_key_event (aglbackend, (PgmEventKey *) pgmevent, event);
       break;
 
     case kEventRawKeyRepeat:
-      GST_WARNING_OBJECT(NULL, "window_event_handler :: kEventRawKeyRepeat");
+      GST_LOG("window_event_handler :: kEventRawKeyRepeat");
       pgmevent = pgm_event_new (PGM_KEY_PRESS);
       translate_key_event (aglbackend, (PgmEventKey *) pgmevent, event);
       break;
 
     case kEventRawKeyUp:
-      GST_WARNING_OBJECT(NULL, "window_event_handler :: kEventRawKeyUp");
+      GST_LOG("window_event_handler :: kEventRawKeyUp");
       pgmevent = pgm_event_new (PGM_KEY_RELEASE);
       translate_key_event (aglbackend, (PgmEventKey *) pgmevent, event);
       break;
@@ -359,7 +358,7 @@ window_event_handler(EventHandlerCallRef handler,
     break;
 
   case kEventClassWindow:
-    GST_WARNING_OBJECT(NULL, "window_event_handler :: kEventClassWindow");
+    GST_LOG("window_event_handler :: kEventClassWindow");
     GetEventParameter(event, kEventParamDirectObject, typeWindowRef, NULL,
                       sizeof(WindowRef), NULL, &window);
     switch (kind) {
@@ -456,7 +455,7 @@ install_window_event_handler(PgmBackend *backend)
 
   GST_DEBUG_OBJECT (aglbackend, "install_window_event_handler");
   CHECK_MAIN_EVENT_LOOP();
-  GST_WARNING("!! numevents: c: %ld, m: %ld",
+  GST_LOG("!! numevents: c: %ld, m: %ld",
               GetNumEventsInQueue(GetCurrentEventQueue()),
               GetNumEventsInQueue(GetMainEventQueue()));
 
@@ -467,7 +466,7 @@ install_window_event_handler(PgmBackend *backend)
                                   GetEventTypeCount (list), list,
                                   (void*) aglbackend, &ref); // add event handler
 
-  GST_DEBUG_OBJECT(aglbackend, "InstallWindowEventHandler() = %ld", err);
+  GST_LOG_OBJECT(aglbackend, "InstallWindowEventHandler() = %ld", err);
   return TRUE;
 }
 
@@ -477,13 +476,13 @@ post_create_window_event(PgmBackend *backend)
   OSStatus err;
   EventRef windowEvent;
 
-  GST_WARNING_OBJECT(backend, "posting synthetic event.");
+  GST_DEBUG_OBJECT(backend, "posting synthetic event.");
   err = MacCreateEvent(nil, kEventClassCommand, kEventProcessCommand, 0,
                        kEventAttributeNone, &windowEvent);
   if (err == noErr) {
     HICommand command; // set HI command parameter...
 
-    GST_WARNING_OBJECT(backend, "posting synthetic event..");
+    GST_LOG_OBJECT(backend, "posting synthetic event..");
 
     command.commandID = kHICommandNew; command.attributes = 0;
     command.menu.menuRef = 0; command.menu.menuItemIndex = 0;
@@ -491,7 +490,7 @@ post_create_window_event(PgmBackend *backend)
     err = SetEventParameter(windowEvent, kEventParamDirectObject,
                             kEventParamHICommand, sizeof(command), &command);
     if (err == noErr) {
-      GST_WARNING_OBJECT(backend, "posting synthetic event...");
+      GST_LOG_OBJECT(backend, "posting synthetic event...");
       err = PostEventToQueue(GetCurrentEventQueue(), windowEvent,
                              kEventPriorityHigh);
     }
@@ -649,24 +648,33 @@ pgm_agl_backend_set_title (PgmBackend *backend, const gchar *title)
    */
 }
 
-static void
-pgm_agl_backend_show (PgmBackend *backend)
+static gboolean
+pgm_agl_backend_set_visibility (PgmBackend *backend,
+                                gboolean visible)
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "show");
+  GST_DEBUG_OBJECT (backend, "set_visibility");
 
-  ShowWindow(aglbackend->win);
+  if (visible)
+    ShowWindow(aglbackend->win);
+  else
+    HideWindow(aglbackend->win);
+
+  return TRUE;
 }
 
-static void
-pgm_agl_backend_hide (PgmBackend *backend)
+static gboolean
+pgm_agl_backend_set_decorated (PgmBackend *backend,
+                               gboolean decorated)
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "hide");
+  GST_DEBUG_OBJECT (backend, "set_decorated");
 
-  HideWindow(aglbackend->win);
+  /* FIXME */
+
+  return FALSE;
 }
 
 static void
@@ -674,7 +682,7 @@ pgm_agl_backend_swap_buffers (PgmBackend *backend)
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "swap_buffers");
+  GST_LOG_OBJECT (backend, "swap_buffers");
 
   aglSwapBuffers(aglbackend->ctx);
 }
@@ -685,7 +693,7 @@ pgm_agl_backend_get_proc_address (PgmBackend *backend,
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "get_proc_address");
+  GST_LOG_OBJECT (backend, "get_proc_address");
 
 
   return CFBundleGetFunctionPointerForName(aglbackend->gl_bundle_ref,
@@ -741,7 +749,7 @@ pgm_agl_backend_get_screen_size_mm (PgmBackend *backend,
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "get_screen_size_mm");
+  GST_LOG_OBJECT (backend, "get_screen_size_mm");
 
   *width = aglbackend->size_mm_width;
   *height = aglbackend->size_mm_height;
@@ -764,7 +772,7 @@ pgm_agl_backend_get_screen_resolution (PgmBackend *backend,
 {
   PgmAglBackend *aglbackend = PGM_AGL_BACKEND(backend);
 
-  GST_DEBUG_OBJECT (backend, "get_screen_resolution");
+  GST_LOG_OBJECT (backend, "get_screen_resolution");
 
   *width = aglbackend->resolution_width;
   *height = aglbackend->resolution_height;
@@ -785,7 +793,7 @@ pgm_agl_backend_build_text_lists (PgmBackend *backend)
   aglbackend->text_lists = gl->gen_lists(256);
   if (!gl->is_list(aglbackend->text_lists))
   {
-    GST_ERROR_OBJECT(aglbackend, "unable to build text display lists\n");
+    GST_WARNING_OBJECT(aglbackend, "unable to build text display lists\n");
     return FALSE;
   }
 
@@ -844,7 +852,13 @@ pgm_agl_backend_raster_text (PgmBackend *backend,
 static void
 pgm_agl_backend_wait_for_vblank (PgmBackend *backend)
 {
-  GST_DEBUG_OBJECT (backend, "wait_for_vblank");
+  GST_LOG_OBJECT (backend, "wait_for_vblank");
+}
+
+static void
+pgm_agl_backend_notify_startup_complete (PgmBackend *backend)
+{
+  GST_LOG_OBJECT (backend, "notify_startup_complete");
 }
 
 static gboolean
@@ -865,6 +879,13 @@ pgm_agl_backend_set_icon (PgmBackend *backend,
   return FALSE;
 }
 
+static void
+pgm_agl_backend_set_drag_status (PgmBackend *backend,
+                                 gboolean accept)
+{
+  GST_LOG_OBJECT (backend, "set_drag_status");
+}
+
 static gboolean
 pgm_agl_backend_is_accelerated (PgmBackend *backend)
 {
@@ -873,9 +894,45 @@ pgm_agl_backend_is_accelerated (PgmBackend *backend)
   return FALSE;
 }
 
+static gboolean
+pgm_agl_backend_is_embeddable (PgmBackend *backend)
+{
+  PgmAglBackend *aglbackend = PGM_AGL_BACKEND (backend);
+
+  GST_LOG_OBJECT (backend, "is_embeddable");
+
+  /* Not supported yet */
+
+  return FALSE;
+}
+
+static void
+pgm_agl_backend_get_embedding_id (PgmBackend *backend,
+                                  gulong *embedding_id)
+{
+  PgmAglBackend *aglbackend = PGM_AGL_BACKEND (backend);
+
+  GST_LOG_OBJECT (backend, "get_embedding_id");
+
+  *embedding_id = 0;
+}
+
+static gboolean
+pgm_agl_backend_has_alpha_component (PgmBackend *backend)
+{
+  GST_DEBUG_OBJECT (backend, "has_alpha_component");
+
+  return FALSE;
+}
+
 /* GObject stuff */
 
-G_DEFINE_TYPE (PgmAglBackend, pgm_agl_backend, PGM_TYPE_BACKEND);
+PGM_DEFINE_DYNAMIC_TYPE (PgmAglBackend, pgm_agl_backend, PGM_TYPE_BACKEND);
+
+void pgm_agl_backend_register (GTypeModule *module)
+{
+  pgm_agl_backend_register_type (module);
+}
 
 static void
 pgm_agl_backend_dispose (GObject *object)
@@ -902,6 +959,9 @@ pgm_agl_backend_class_init (PgmAglBackendClass *klass)
   GObjectClass *gobject_class;
   PgmBackendClass *backend_class;
 
+  GST_DEBUG_CATEGORY_INIT (pgm_gl_agl_debug, "pgm_gl_aglbackend", 0,
+                           "OpenGL plugin: PgmAglBackend");
+
   parent_class = g_type_class_peek_parent (klass);
 
   gobject_class = G_OBJECT_CLASS (klass);
@@ -916,8 +976,8 @@ pgm_agl_backend_class_init (PgmAglBackendClass *klass)
   backend_class->destroy_window =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_destroy_window);
   backend_class->set_title = GST_DEBUG_FUNCPTR (pgm_agl_backend_set_title);
-  backend_class->show = GST_DEBUG_FUNCPTR (pgm_agl_backend_show);
-  backend_class->hide = GST_DEBUG_FUNCPTR (pgm_agl_backend_hide);
+  backend_class->set_decorated =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_set_decorated);
   backend_class->swap_buffers =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_swap_buffers);
   backend_class->get_proc_address =
@@ -925,6 +985,8 @@ pgm_agl_backend_class_init (PgmAglBackendClass *klass)
   backend_class->set_size = GST_DEBUG_FUNCPTR (pgm_agl_backend_set_size);
   backend_class->set_fullscreen =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_set_fullscreen);
+  backend_class->set_visibility =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_set_visibility);
   backend_class->get_screen_size_mm =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_get_screen_size_mm);
   backend_class->set_screen_resolution =
@@ -938,16 +1000,32 @@ pgm_agl_backend_class_init (PgmAglBackendClass *klass)
   backend_class->raster_text = GST_DEBUG_FUNCPTR (pgm_agl_backend_raster_text);
   backend_class->wait_for_vblank =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_wait_for_vblank);
+  backend_class->notify_startup_complete =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_notify_startup_complete);
   backend_class->set_cursor = GST_DEBUG_FUNCPTR (pgm_agl_backend_set_cursor);
   backend_class->set_icon = GST_DEBUG_FUNCPTR (pgm_agl_backend_set_icon);
+  backend_class->set_drag_status =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_set_drag_status);
   backend_class->is_accelerated =
     GST_DEBUG_FUNCPTR (pgm_agl_backend_is_accelerated);
+  backend_class->is_embeddable =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_is_embeddable);
+  backend_class->get_embedding_id =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_get_embedding_id);
+  backend_class->has_alpha_component =
+    GST_DEBUG_FUNCPTR (pgm_agl_backend_has_alpha_component);
+}
+
+static void
+pgm_agl_backend_class_finalize (PgmAglBackendClass *klass)
+{
+  return;
 }
 
 static void
 pgm_agl_backend_init (PgmAglBackend *aglbackend)
 {
-  GST_DEBUG_OBJECT (aglbackend, "init");
+  GST_INFO_OBJECT (aglbackend, "init");
 
   aglbackend->win = NULL;
   aglbackend->pixfmt = NULL;
