@@ -2,7 +2,7 @@
  *
  * Pigment OpenGL plugin
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -48,26 +48,40 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-#include <GL/gl.h>
+#include "pgmcontext.h"
+
+#ifdef G_OS_UNIX
 #include <unistd.h> /* pipe, close */
+#endif /* G_OS_UNIX */
+
+#ifdef WIN32
+#include <windows.h>
+#include <fcntl.h>    /* _O_BINARY */
+#include <io.h>       /* _pipe, _close */
+#define pipe(handle)  _pipe(handle, 4096, _O_BINARY)
+#define close(handle) _close(handle)
+#endif /* WIN32 */
+
+#include <GL/gl.h>  /* gl* */
 #include <string.h> /* strlen */
 #include <ctype.h>  /* isdigit */
 #include <math.h>   /* tan */
 
-#include "pgmcontext.h"
-#ifdef G_OS_UNIX
 #ifdef HAVE_AGL_AGL_H
 #include "pgmaglbackend.h"
-#else
+#elif defined HAVE_GL_GLX_H
 #include "pgmglxbackend.h"
-#endif /* HAVE_AGL_AGL_H */
-#elif defined G_OS_WIN32
+#elif defined HAVE_WGL_H
 #include "pgmwglbackend.h"
-#endif /* G_OS_UNIX */
+#endif
 
 GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
 #define GST_CAT_DEFAULT pgm_gl_debug
 
+/* Function prototypes */
+static void flush_task_queue (PgmContext *context, GList **queue);
+static void render           (PgmContext *context);
+
 /* Map OpenGL extension names and feature bits */
 typedef struct {
   const gchar *name;
@@ -93,6 +107,7 @@ static ExtensionMap gl_extensions_map[] = {
   { "GL_ARB_vertex_buffer_object",     PGM_GL_FEAT_VERTEX_BUFFER_OBJECT     },
   { "GL_ARB_pixel_buffer_object",      PGM_GL_FEAT_PIXEL_BUFFER_OBJECT      },
   { "GL_EXT_pixel_buffer_object",      PGM_GL_FEAT_PIXEL_BUFFER_OBJECT      },
+  { "GL_EXT_blend_func_separate",      PGM_GL_FEAT_BLEND_FUNC_SEPARATE      },
   { "GL_EXT_blend_color",              PGM_GL_FEAT_BLEND_COLOR              },
   { "GL_ARB_imaging",                  PGM_GL_FEAT_BLEND_COLOR              },
   { "GL_APPLE_packed_pixels",          PGM_GL_FEAT_PACKED_PIXELS            },
@@ -114,6 +129,7 @@ static PgmContextProcAddress gl_proc_address = {
   (pgm_gl_enable_client_state)   glEnableClientState,
   (pgm_gl_disable_client_state)  glDisableClientState,
   (pgm_gl_vertex_pointer)        glVertexPointer,
+  (pgm_gl_color_pointer)         glColorPointer,
   (pgm_gl_tex_coord_pointer)     glTexCoordPointer,
   (pgm_gl_draw_arrays)           glDrawArrays,
   (pgm_gl_tex_env_f)             glTexEnvf,
@@ -179,6 +195,7 @@ static PgmContextProcAddress gl_proc_address = {
   (pgm_gl_get_float_v)           glGetFloatv,
 
   /* OpenGL 1.2 extension functions */
+  (pgm_gl_blend_func_separate)           NULL,
   (pgm_gl_blend_color)                   NULL,
   (pgm_gl_active_texture)                NULL,
   (pgm_gl_client_active_texture)         NULL,
@@ -219,7 +236,7 @@ update_projection (PgmContext *context)
   PgmContextProcAddress *gl = context->gl;
   gint projected_width, projected_height;
   gint projected_x, projected_y;
-  PgmMat4x4 projection, transpose;
+  PgmMat4x4 *projection, *transpose;
 
   /* Retrieve viewport informations with lock */
   GST_OBJECT_LOCK (viewport);
@@ -233,12 +250,12 @@ update_projection (PgmContext *context)
   /* Set up viewport */
   gl->viewport (projected_x, projected_y, projected_width, projected_height);
 
-  /* Update projection matrix, needs the transpose projection since OpenGL uses
-   * column major order */
-  pgm_mat4x4_transpose (&transpose, &projection);
+  /* Update projection matrix */
   gl->matrix_mode (PGM_GL_PROJECTION);
   gl->load_identity ();
-  gl->load_matrix_f (&transpose.m[0][0]);
+  transpose = pgm_mat4x4_transpose (projection);
+  gl->load_matrix_f (transpose->m);
+  pgm_mat4x4_free (transpose);
   gl->matrix_mode (PGM_GL_MODELVIEW);
 
   /* Update all the drawables inside the viewport */
@@ -261,13 +278,21 @@ init_opengl_states (PgmContext *context)
 {
   PgmContextProcAddress *gl = context->gl;
 
+  /* Enable texture mapping */
   gl->enable (PGM_GL_TEXTURE_2D);
-  gl->enable (PGM_GL_DEPTH_TEST);
-  gl->enable (PGM_GL_BLEND);
 
-  gl->depth_func (PGM_GL_LEQUAL);
-  gl->blend_func (PGM_GL_SRC_ALPHA, PGM_GL_ONE_MINUS_SRC_ALPHA);
-  gl->clear_color (0.0f, 0.0f, 0.0f, 1.0f);
+  /* Disable depth testing */
+  gl->disable (PGM_GL_DEPTH_TEST);
+
+  /* Set up standard blending with a different alpha function to correctly
+   * support compositing with the desktop when the alpha clearing color is
+   * available (ARGB visual) */
+  if (context->feature_mask & PGM_GL_FEAT_BLEND_FUNC_SEPARATE)
+    gl->blend_func_separate (PGM_GL_SRC_ALPHA, PGM_GL_ONE_MINUS_SRC_ALPHA,
+                             PGM_GL_SRC_ALPHA, PGM_GL_ONE);
+  else
+    gl->blend_func (PGM_GL_SRC_ALPHA, PGM_GL_ONE_MINUS_SRC_ALPHA);
+  gl->enable (PGM_GL_BLEND);
 }
 
 /* Check whether an extension is supported by the OpenGL implementation given
@@ -277,7 +302,7 @@ has_opengl_extension (const gchar *extensions,
                       const gchar *extension_name)
 {
   gchar *end;
-  gint ext_name_len = strlen (extension_name);
+  size_t ext_name_len = strlen (extension_name);
   gchar *p = (gchar*) extensions;
   size_t size;
 
@@ -324,6 +349,21 @@ bind_opengl_extensions (PgmContext *context)
 {
   PgmContextProcAddress *gl = context->gl;
 
+  /* Blend func separate */
+  if (context->feature_mask & PGM_GL_FEAT_BLEND_FUNC_SEPARATE)
+    {
+      if (context->version >= 1.4f)
+        gl->blend_func_separate = (pgm_gl_blend_func_separate)
+          pgm_backend_get_proc_address (context->backend,
+                                        "glBlendFuncSeparate");
+      else
+        gl->blend_func_separate = (pgm_gl_blend_func_separate)
+          pgm_backend_get_proc_address (context->backend,
+                                        "glBlendFuncSeparateEXT");
+      if (!gl->blend_func_separate)
+        context->feature_mask &= ~PGM_GL_FEAT_BLEND_FUNC_SEPARATE;
+    }
+
   /* Blend color */
   if (context->feature_mask & PGM_GL_FEAT_BLEND_COLOR)
     {
@@ -496,49 +536,48 @@ bind_opengl_extensions (PgmContext *context)
     }
 }
 
-/* Rasterize the FPS string at the top-left of the window */
-G_GNUC_UNUSED static void
-raster_fps_string (PgmContext *context,
-                   gfloat refresh_rate)
+/* Compute the frame rate per second */
+static void
+compute_frame_rate (PgmContext *context)
 {
   GTimeVal current_time;
-  gchar fps_string[16];
   gfloat elapsed_time;
-  static gint fps;
+  static guint fps = 0;
 
-  /* Retrieve current time and compute the time elapsed since the refresh */
+  /* Retrieve current time and compute the time elapsed since the last tick */
   g_get_current_time (&current_time);
-  elapsed_time = (current_time.tv_sec - context->fps_print_time.tv_sec)
-    + (current_time.tv_usec - context->fps_print_time.tv_usec) * 0.000001f;
+  elapsed_time = (current_time.tv_sec - context->fps_tick_time.tv_sec)
+    + (current_time.tv_usec - context->fps_tick_time.tv_usec) * 0.000001f;
 
-  /* Is the refresh_rate elapsed? */
-  if (elapsed_time > refresh_rate)
+  /* Is the second elapsed? */
+  if (elapsed_time >= 1.0f)
     {
-      /* Compute the time elapsed since the last update */
-      elapsed_time = (current_time.tv_sec - context->update_time.tv_sec)
-        + (current_time.tv_usec - context->update_time.tv_usec) * 0.000001f;
-      fps = 1.0f / elapsed_time;
+      context->fps = fps;
 
-      /* Set the new FPS print time */
-      context->fps_print_time.tv_sec = current_time.tv_sec;
-      context->fps_print_time.tv_usec = current_time.tv_usec;
+      /* Set the new tick time */
+      context->fps_tick_time.tv_sec = current_time.tv_sec;
+      context->fps_tick_time.tv_usec = current_time.tv_usec;
+
+      /* Reset the counter */
+      fps = 0;
     }
 
-  /* Build and rasterize the FPS string */
-  g_snprintf (fps_string, 16, "%d FPS", fps);
+  fps++;
 
-  context->gl->disable (PGM_GL_TEXTURE_2D);
-  context->gl->disable (PGM_GL_DEPTH_TEST);
-  pgm_backend_raster_text (context->backend, fps_string, 6.0f, 15.0f,
-                           0.0f, 0.0f, 0.0f);
-  pgm_backend_raster_text (context->backend, fps_string, 5.0f, 14.0f,
-                           1.0f, 1.0f, 1.0f);
-  context->gl->enable (PGM_GL_TEXTURE_2D);
-  context->gl->enable (PGM_GL_DEPTH_TEST);
+#if 0 /* FIXME: Broken */
+  {
+    gchar fps_string[16];
 
-  /* Set the new update time */
-  context->update_time.tv_sec = current_time.tv_sec;
-  context->update_time.tv_usec = current_time.tv_usec;
+    /* Build and rasterize the FPS string */
+    g_snprintf (fps_string, 16, "%d FPS", fps);
+    context->gl->disable (PGM_GL_TEXTURE_2D);
+    pgm_backend_raster_text (context->backend, fps_string, 6.0f, 15.0f,
+                             0.0f, 0.0f, 0.0f);
+    pgm_backend_raster_text (context->backend, fps_string, 5.0f, 14.0f,
+                             1.0f, 1.0f, 1.0f);
+    context->gl->enable (PGM_GL_TEXTURE_2D);
+  }
+#endif
 }
 
 /* Viewport update task handler */
@@ -547,6 +586,7 @@ do_projection (PgmContext *context,
                gpointer data)
 {
   update_projection (context);
+  pgm_context_update (context);
 }
 
 /* Size update task handler */
@@ -562,6 +602,7 @@ do_size (PgmContext *context,
   GST_OBJECT_UNLOCK (context->glviewport);
 
   pgm_backend_set_size (context->backend, width, height);
+  pgm_context_update (context);
 }
 
 /* Title update task handler */
@@ -580,6 +621,20 @@ do_title (PgmContext *context,
   g_free (title);
 }
 
+/* Decoration update task handler */
+static void
+do_decoration (PgmContext *context,
+               gpointer data)
+{
+  gboolean decorated;
+
+  GST_OBJECT_LOCK (context->glviewport);
+  decorated = PGM_VIEWPORT (context->glviewport)->decorated;
+  GST_OBJECT_UNLOCK (context->glviewport);
+
+  pgm_backend_set_decorated (context->backend, decorated);
+}
+
 /* Fullscreen update task handler */
 static void
 do_fullscreen (PgmContext *context,
@@ -592,6 +647,26 @@ do_fullscreen (PgmContext *context,
   GST_OBJECT_UNLOCK (context->glviewport);
 
   pgm_backend_set_fullscreen (context->backend, fullscreen);
+  pgm_context_update (context);
+}
+
+/* Alpha blending update task handler */
+static void
+do_alpha_blending (PgmContext *context,
+                   gpointer data)
+{
+  gboolean alpha_blending;
+
+  GST_OBJECT_LOCK (context->glviewport);
+  alpha_blending = PGM_VIEWPORT (context->glviewport)->alpha_blending;
+  GST_OBJECT_UNLOCK (context->glviewport);
+
+  if (alpha_blending)
+    context->gl->enable (PGM_GL_BLEND);
+  else
+    context->gl->disable (PGM_GL_BLEND);
+
+  pgm_context_update (context);
 }
 
 /* Resolution update task handler */
@@ -602,6 +677,46 @@ do_resolution (PgmContext *context,
   /* FIXME */
 }
 
+/* Opacity task handler */
+static void
+do_opacity (PgmContext *context,
+            gpointer data)
+{
+  gfloat opacity;
+
+  GST_OBJECT_LOCK (context->glviewport);
+  opacity = PGM_VIEWPORT (context->glviewport)->opacity * INV_255;
+  GST_OBJECT_UNLOCK (context->glviewport);
+
+  context->gl->clear_color (0.0f, 0.0f, 0.0f, opacity);
+  pgm_context_update (context);
+}
+
+/* Visibility update task handler */
+static void
+do_visibility (PgmContext *context,
+               gpointer data)
+{
+  static gboolean first = TRUE;
+  gboolean visible;
+
+  GST_OBJECT_LOCK (context->glviewport);
+  visible = PGM_VIEWPORT (context->glviewport)->visible;
+  GST_OBJECT_UNLOCK (context->glviewport);
+
+  /* First time window is shown */
+  if (G_UNLIKELY (first && visible))
+    {
+      /* The clear color needs an update, to avoid flashes */
+      do_opacity (context, NULL);
+
+      pgm_backend_notify_startup_complete (context->backend);
+      first = FALSE;
+    }
+
+  pgm_backend_set_visibility (context->backend, visible);
+}
+
 /* Cursor update task handler */
 static void
 do_cursor (PgmContext *context,
@@ -630,6 +745,67 @@ do_icon (PgmContext *context,
   pgm_backend_set_icon (context->backend, icon);
 }
 
+/* Drag status update task handler */
+static void
+do_drag_status (PgmContext *context,
+                gpointer data)
+{
+  pgm_backend_set_drag_status (context->backend,
+                               context->glviewport->drag_status);
+}
+
+/* Frame buffer read task handler */
+static void
+do_read_pixels (PgmContext *context,
+                gpointer data)
+{
+  PgmGlViewportPixelRectangle *rectangle = (PgmGlViewportPixelRectangle*) data;
+  PgmContextProcAddress *gl = context->gl;
+  PgmViewport *viewport = PGM_VIEWPORT (context->glviewport);
+  PgmCanvas *canvas = viewport->canvas;
+
+  /* Flush the queues ensuring last rendering requests will be completed */
+  pgm_gl_viewport_flush_update_queue (context->glviewport);
+  flush_task_queue (context, &context->immediate_task);
+  flush_task_queue (context, &context->deferred_task);
+
+  /* If there's a canvas we need to invert the projection matrix upside-down.
+   * Note that if there's no canvas bound, the pixels area will be black. */
+  if (G_LIKELY (canvas))
+    {
+      gl->matrix_mode (PGM_GL_PROJECTION);
+      gl->push_matrix ();
+      gl->scale_f (1.0f, -1.0f, 1.0f);
+      gl->translate_f (0.0f, -canvas->height, 0.0f);
+      gl->matrix_mode (PGM_GL_MODELVIEW);
+    }
+
+  /* Draw the scene, read it from the back buffer, and clear the back buffer */
+  render (context);
+  gl->read_buffer (PGM_GL_BACK);
+  gl->read_pixels (rectangle->x, rectangle->y, rectangle->width,
+                   rectangle->height, GL_RGBA, GL_UNSIGNED_BYTE,
+                   rectangle->pixels);
+  gl->clear (PGM_GL_COLOR_BUFFER_BIT);
+  gl->read_buffer (PGM_GL_FRONT);
+
+  /* Pop our inverted projection matrix if needed */
+  if (G_LIKELY (canvas))
+    {
+      context->gl->matrix_mode (PGM_GL_PROJECTION);
+      context->gl->pop_matrix ();
+      context->gl->matrix_mode (PGM_GL_MODELVIEW);
+    }
+
+  /* Push the read pixels back to the viewport */
+  pgm_viewport_push_pixels (viewport, rectangle->width, rectangle->height,
+                            rectangle->pixels);
+
+  /* Then finally free the rectangle structure */
+  g_slice_free (PgmGlViewportPixelRectangle, rectangle);
+  rectangle = NULL;
+}
+
 /* Texture generation task handler */
 static void
 do_gen_texture (PgmContext *context,
@@ -678,11 +854,10 @@ render (PgmContext *context)
 
   g_mutex_lock (glviewport->layer_lock);
 
-  g_list_foreach (glviewport->far, (GFunc) pgm_gl_drawable_draw, NULL);
-  context->gl->clear (PGM_GL_DEPTH_BUFFER_BIT);
-  g_list_foreach (glviewport->middle, (GFunc) pgm_gl_drawable_draw, NULL);
-  context->gl->clear (PGM_GL_DEPTH_BUFFER_BIT);
-  g_list_foreach (glviewport->near, (GFunc) pgm_gl_drawable_draw, NULL);
+  g_list_foreach (glviewport->far_layer, (GFunc) pgm_gl_drawable_draw, NULL);
+  g_list_foreach (glviewport->middle_layer,
+                  (GFunc) pgm_gl_drawable_draw, NULL);
+  g_list_foreach (glviewport->near_layer, (GFunc) pgm_gl_drawable_draw, NULL);
 
   g_mutex_unlock (glviewport->layer_lock);
 }
@@ -811,9 +986,9 @@ flush_task_queue (PgmContext *context,
 
 /* i/o watch callback for immediate tasks */
 static gboolean
-immediate_cb (GIOChannel *source,
-              GIOCondition condition,
-              gpointer data)
+immediate_io_cb (GIOChannel *source,
+                 GIOCondition condition,
+                 gpointer data)
 {
   PgmContext *context = PGM_CONTEXT (data);
   gchar buf;
@@ -826,29 +1001,67 @@ immediate_cb (GIOChannel *source,
   return TRUE;
 }
 
-/* i/o watch callback for updates */
+/* Update source */
 static gboolean
-update_cb (GIOChannel *source,
-           GIOCondition condition,
-           gpointer data)
+update_cb (gpointer data)
 {
   PgmContext *context = PGM_CONTEXT (data);
-  gchar buf;
-
-  g_io_channel_read_chars (source, &buf, 1, NULL, NULL);
 
   /* Flush update queue */
   pgm_gl_viewport_flush_update_queue (context->glviewport);
 
-  /* Complete context tasks */
+  /* Flush rendering queues */
   flush_task_queue (context, &context->immediate_task);
   flush_task_queue (context, &context->deferred_task);
 
-  /* And refresh the scene */
+  /* Render the scene */
   render (context);
   pgm_backend_wait_for_vblank (context->backend);
   pgm_backend_swap_buffers (context->backend);
-  context->gl->clear (PGM_GL_COLOR_BUFFER_BIT | PGM_GL_DEPTH_BUFFER_BIT);
+  context->gl->clear (PGM_GL_COLOR_BUFFER_BIT);
+
+  /* Handle frame rate */
+  compute_frame_rate (context);
+
+  return TRUE;
+}
+
+/* Update timeout used to remove the update idle after 1 second of inactivity */
+static gboolean
+update_removal_timeout_cb (gpointer data)
+{
+  PgmContext *context = (PgmContext*) data;
+  GTimeVal current_time;
+
+  g_get_current_time (&current_time);
+
+  g_mutex_lock (context->update_mutex);
+
+  /* If the elapsed time since the previous update request rises above one
+   * second, remove the auto-update source */
+  if ((current_time.tv_sec - context->update_timestamp.tv_sec) >= 1)
+    {
+      GSource *source;
+
+      /* Get the source from the rendering GMainContext and remove it */
+      source = g_main_context_find_source_by_id (context->render_context,
+                                                 context->update_tag);
+      if (source)
+        g_source_destroy (source);
+
+      context->update_tag = 0;
+      context->auto_updated = FALSE;
+      context->fps = 0;
+
+      g_mutex_unlock (context->update_mutex);
+
+      GST_DEBUG ("removing update source");
+
+      /* Remove ourself from the loop */
+      return FALSE;
+    }
+
+  g_mutex_unlock (context->update_mutex);
 
   return TRUE;
 }
@@ -937,45 +1150,48 @@ render_loop (gpointer data)
 {
   PgmContext *context = (PgmContext *) data;
   PgmContextProcAddress *gl = NULL;
+  const gchar *env_var = NULL;
   gint i = 0;
 
   /* Instantiate correct backend depending on the system at compile-time */
-#ifdef G_OS_UNIX
-#ifdef HAVE_AGL_AGL_H
-  context->backend = PGM_BACKEND (pgm_agl_backend_new (context));
-#else
+#if defined (HAVE_GL_GLX_H)
   context->backend = PGM_BACKEND (pgm_glx_backend_new (context));
-#endif /* HAVE_AGL_AGL_H */
-#elif defined G_OS_WIN32
+#elif defined (HAVE_AGL_AGL_H)
+  context->backend = PGM_BACKEND (pgm_agl_backend_new (context));
+#elif defined (HAVE_WGL_H)
   context->backend = PGM_BACKEND (pgm_wgl_backend_new (context));
-#endif /* G_OS_UNIX */
+#endif
 
   pgm_backend_create_window (context->backend);
 
   context->gl = gl = &gl_proc_address;
 
-  /* Retrieve OpenGL version */
+  /* Retrieve renderer informations */
   context->version_string = (const gchar *) gl->get_string (PGM_GL_VERSION);
+  context->vendor = (const gchar *) gl->get_string (PGM_GL_VENDOR);
+  context->renderer = (const gchar *) gl->get_string (PGM_GL_RENDERER);
+  context->extensions = (const gchar *) gl->get_string (PGM_GL_EXTENSIONS);
+
+  /* Add some debug info */
+  GST_INFO ("OpenGL vendor: %s", context->vendor);
+  GST_INFO ("OpenGL renderer: %s", context->renderer);
+  GST_INFO ("OpenGL version: %s", context->version_string);
+  GST_DEBUG ("OpenGL extensions: %s", context->extensions);
+
+  /* Check that OpenGL minimum version is 1.2 */
   if (!context->version_string)
     {
       GST_ERROR ("couldn't retrieve OpenGL version");
-      exit (1);
+      exit (1); /* FIXME: So bad!!! */
     }
-
-  /* Check that OpenGL minimum version is 1.2 */
   context->version = get_float_opengl_version (context->version_string);
   if (context->version < 1.2f)
     {
       GST_ERROR ("OpenGL version %.1f (1.2 minimum required)",
                  context->version);
-      exit (1);
+      exit (1); /* FIXME: So bad!!! */
     }
 
-  /* Retrieve renderer informations */
-  context->vendor = (const gchar *) gl->get_string (PGM_GL_VENDOR);
-  context->renderer = (const gchar *) gl->get_string (PGM_GL_RENDERER);
-  context->extensions = (const gchar *) gl->get_string (PGM_GL_EXTENSIONS);
-
   /* Retrieve and bind supported extensions */
   while (gl_extensions_map[i].name)
     {
@@ -983,47 +1199,83 @@ render_loop (gpointer data)
         context->feature_mask |= gl_extensions_map[i].feature_mask;
       i++;
     }
+
+  /* Check in the environment if we have to disable ARB vp/fp */
+  if (context->feature_mask & PGM_GL_FEAT_FRAGMENT_PROGRAM)
+    {
+      GST_INFO ("OpenGL fragment programs are supported");
+      env_var = g_getenv ("PGM_GL_PROGRAMS");
+      if (env_var && '0' == env_var[0])
+        {
+          GST_INFO ("Deactivating ARB vertex and fragment programs");
+          context->feature_mask &= ~PGM_GL_FEAT_FRAGMENT_PROGRAM;
+        }
+    }
+  else
+    GST_INFO ("OpenGL fragment programs are not supported");
+
   bind_opengl_extensions (context);
 
   /* Maximum POT and NPOT texture size */
   gl->get_integer_v (PGM_GL_MAX_TEXTURE_SIZE, &context->max_texture_2d_size);
+  GST_INFO ("OpenGL max texture 2D size: %d", context->max_texture_2d_size);
   if (context->feature_mask & PGM_GL_FEAT_TEXTURE_RECTANGLE)
-    gl->get_integer_v (PGM_GL_MAX_RECTANGLE_TEXTURE_SIZE,
-                       &context->max_texture_rect_size);
+    {
+      gl->get_integer_v (PGM_GL_MAX_RECTANGLE_TEXTURE_SIZE,
+                         &context->max_texture_rect_size);
+      GST_INFO ("OpenGL max texture rect size: %d",
+                context->max_texture_rect_size);
+    }
 
   /* Pixel store default values */
   gl->get_integer_v (PGM_GL_UNPACK_ROW_LENGTH, &context->row_length);
   gl->get_integer_v (PGM_GL_UNPACK_SKIP_ROWS, &context->skip_rows);
   gl->get_integer_v (PGM_GL_UNPACK_SKIP_PIXELS, &context->skip_pixels);
 
-  /* Texture units size */
-  if (context->feature_mask & PGM_GL_FEAT_FRAGMENT_PROGRAM
-      && context->feature_mask & PGM_GL_FEAT_MULTITEXTURE)
+  /* Multi texturing */
+  if (context->feature_mask & PGM_GL_FEAT_MULTITEXTURE)
     {
       gl->get_integer_v (PGM_GL_MAX_TEXTURE_IMAGE_UNITS,
                          &context->max_texture_units);
-      if (context->max_texture_units >= 3)
-        context->feature_mask |= PGM_GL_FEAT_PER_PLANE_YCBCR_PROGRAM;
+      GST_INFO ("OpenGL multitexturing is supported, %d image units",
+                context->max_texture_units);
+      if (context->feature_mask & PGM_GL_FEAT_FRAGMENT_PROGRAM
+          && context->max_texture_units >= 3)
+        {
+          GST_INFO ("OpenGL shader for color space conversions is available");
+          context->feature_mask |= PGM_GL_FEAT_PER_PLANE_YCBCR_PROGRAM;
+        }
     }
+  else
+    GST_INFO ("OpenGL multitexturing is not supported");
 
   /* Rasterized text display lists */
   pgm_backend_build_text_lists (context->backend);
 
   /* Init OpenGL */
   init_opengl_states (context);
-  gl->clear (PGM_GL_COLOR_BUFFER_BIT | PGM_GL_DEPTH_BUFFER_BIT);
+  gl->clear (PGM_GL_COLOR_BUFFER_BIT);
 
   /* Compile the ARB fragment programs if supported */
   if (context->feature_mask & PGM_GL_FEAT_FRAGMENT_PROGRAM)
     pgm_program_create (context);
 
-  /* Add the event sources to the rendering thread GMainContext */
-  context->update_id = add_input_watch (context->update_out,
-                                        (GIOFunc) update_cb, context,
-                                        context->render_context);
-  context->immediate_id = add_input_watch (context->immediate_out,
-                                           (GIOFunc) immediate_cb, context,
-                                           context->render_context);
+  /* Add the immediate source to the rendering thread GMainContext */
+  context->immediate_tag = add_input_watch (context->immediate_out,
+                                            (GIOFunc) immediate_io_cb, context,
+                                            context->render_context);
+
+  /* Get the framerate requested by the user in the environment */
+  env_var = g_getenv ("PGM_GL_FPS");
+  if (env_var)
+    {
+      gint fps = atoi (env_var);
+      if (fps > 0)
+        {
+          GST_INFO ("Requested FPS: %d", fps);
+          context->requested_fps = 1000 / fps;
+        }
+    }
 
   /* Signal application thread that the rendering thread is now initialized */
   g_mutex_lock (context->init_mutex);
@@ -1068,21 +1320,12 @@ init_context (PgmContext *context)
   context->init_cond = g_cond_new ();
   context->initialized = FALSE;
 
-  /* Update task source */
-  context->update_fd[0] = -1;
-  context->update_fd[1] = -1;
-  context->update_in = NULL;
-  context->update_out = NULL;
-  context->update_id = 0;
-  create_io_channels (context->update_fd, &context->update_in,
-                      &context->update_out);
-
   /* Immediate task source */
   context->immediate_fd[0] = -1;
   context->immediate_fd[1] = -1;
   context->immediate_in = NULL;
   context->immediate_out = NULL;
-  context->immediate_id = 0;
+  context->immediate_tag = 0;
   create_io_channels (context->immediate_fd, &context->immediate_in,
                       &context->immediate_out);
 
@@ -1090,9 +1333,16 @@ init_context (PgmContext *context)
   context->immediate_task = NULL;
   context->deferred_task = NULL;
 
-  /* FPS displaying timers */
-  g_get_current_time (&context->update_time);
-  g_get_current_time (&context->fps_print_time);
+  /* Auto-update */
+  context->update_mutex = g_mutex_new ();
+  context->auto_updated = FALSE;
+  g_get_current_time (&context->update_timestamp);
+  context->update_tag = 0;
+  context->requested_fps = 0;
+
+  /* Frame rate */
+  g_get_current_time (&context->fps_tick_time);
+  context->fps = 0;
 
   /* Create rendering thread */
   context->render_thread = g_thread_create (render_loop, context, TRUE, &error);
@@ -1106,12 +1356,22 @@ init_context (PgmContext *context)
   context->task_func[PGM_CONTEXT_PROJECTION] = GST_DEBUG_FUNCPTR (do_projection);
   context->task_func[PGM_CONTEXT_SIZE] = GST_DEBUG_FUNCPTR (do_size);
   context->task_func[PGM_CONTEXT_TITLE] = GST_DEBUG_FUNCPTR (do_title);
+  context->task_func[PGM_CONTEXT_DECORATION] = GST_DEBUG_FUNCPTR (do_decoration);
   context->task_func[PGM_CONTEXT_FULLSCREEN] =
     GST_DEBUG_FUNCPTR (do_fullscreen);
+  context->task_func[PGM_CONTEXT_VISIBILITY] =
+    GST_DEBUG_FUNCPTR (do_visibility);
+  context->task_func[PGM_CONTEXT_ALPHA_BLENDING] =
+    GST_DEBUG_FUNCPTR (do_alpha_blending);
   context->task_func[PGM_CONTEXT_RESOLUTION] =
     GST_DEBUG_FUNCPTR (do_resolution);
+  context->task_func[PGM_CONTEXT_OPACITY] = GST_DEBUG_FUNCPTR (do_opacity);
   context->task_func[PGM_CONTEXT_CURSOR] = GST_DEBUG_FUNCPTR (do_cursor);
   context->task_func[PGM_CONTEXT_ICON] = GST_DEBUG_FUNCPTR (do_icon);
+  context->task_func[PGM_CONTEXT_DRAG_STATUS] =
+    GST_DEBUG_FUNCPTR (do_drag_status);
+  context->task_func[PGM_CONTEXT_READ_PIXELS] =
+    GST_DEBUG_FUNCPTR (do_read_pixels);
 
   /* Texture related tasks function binding */
   context->task_func[PGM_CONTEXT_GEN_TEXTURE] =
@@ -1142,11 +1402,18 @@ dispose_context (PgmContext *context)
   while (g_main_context_pending (context->render_context))
     g_main_context_iteration (context->render_context, FALSE);
 
-  /* Remove the event sources */
-  if (context->immediate_id)
-    g_source_remove (context->immediate_id);
-  if (context->update_id)
-    g_source_remove (context->update_id);
+  /* Remove the immediate source */
+  if (context->immediate_tag)
+    g_source_remove (context->immediate_tag);
+
+  /* Remove the auto-update source */
+  if (context->update_tag)
+    {
+      GSource *source = g_main_context_find_source_by_id
+        (context->render_context, context->update_tag);
+      if (source)
+        g_source_destroy (source);
+    }
 
   /* Join the rendering thread */
   g_main_loop_quit (context->render_loop);
@@ -1154,16 +1421,17 @@ dispose_context (PgmContext *context)
   g_main_context_unref (context->render_context);
   g_thread_join (context->render_thread);
 
-  /* Dispose event sources data */
+  /* Dispose immediate source data */
   delete_io_channels (context->immediate_fd, &context->immediate_in,
                       &context->immediate_out);
-  delete_io_channels (context->update_fd, &context->update_in,
-                      &context->update_out);
 
   /* Free the task queues */
   free_task_queue (&context->deferred_task);
   free_task_queue (&context->immediate_task);
 
+  /* Auto-update lock */
+  g_mutex_free (context->update_mutex);
+
   /* Initialization lock */
   g_mutex_free (context->init_mutex);
   g_cond_free (context->init_cond);
@@ -1234,7 +1502,46 @@ pgm_context_update (PgmContext *context)
 {
   g_return_if_fail (context != NULL);
 
-  write_char (context->update_in);
+  g_mutex_lock (context->update_mutex);
+
+  /* Get a timestamp of the last update request */
+  g_get_current_time (&context->update_timestamp);
+
+  /* Add the auto-update if the rendering loop is stopped */
+  if (G_UNLIKELY (!context->auto_updated))
+    {
+      GSource *source;
+
+      /* Add the update source to the rendering GMainContext. If the user has
+       * not specified a framerate through the PGM_GL_FPS environment
+       * variable, an idle source blocked on the vsync is attached. Otherwise,
+       * a timeout source with the specified framerate is attached. */
+      if (G_LIKELY (context->requested_fps == 0))
+        {
+          GST_DEBUG ("adding update idle source");
+          source = g_idle_source_new ();
+          g_source_set_priority ((GSource*) source, G_PRIORITY_DEFAULT);
+        }
+      else
+        {
+          GST_DEBUG ("adding update timeout source");
+          source = g_timeout_source_new (context->requested_fps);
+        }
+      g_source_set_callback (source, update_cb, context, NULL);
+      context->update_tag = g_source_attach (source, context->render_context);
+      g_source_unref (source);
+
+      /* Add the update removal timeout to the rendering GMainContext */
+      source = g_timeout_source_new (500);
+      g_source_set_callback (source, update_removal_timeout_cb, context, NULL);
+      g_source_attach (source, context->render_context);
+      g_source_unref (source);
+
+      /* Mark the render loop as auto-updated */
+      context->auto_updated = TRUE;
+    }
+
+  g_mutex_unlock (context->update_mutex);
 }
 
 void
