@@ -2,7 +2,7 @@
  *
  * Pigment OpenGL plugin
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -37,8 +37,42 @@
 #include <math.h>
 #include "pgmgltext.h"
 
-GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
-#define GST_CAT_DEFAULT pgm_gl_debug
+#ifndef HAVE_PANGO_1_16
+#ifndef G_OS_WIN32
+#include <locale.h>
+#endif
+#endif
+
+/* This is the offset such that (0xFF << PGM_n_OFFSET) is the mask to access the
+ * nth byte of a guint32 */
+#if G_BYTE_ORDER == G_LITTLE_ENDIAN
+#define BYTE_1_OFFSET 0
+#define BYTE_2_OFFSET 8
+#define BYTE_3_OFFSET 16
+#define BYTE_4_OFFSET 24
+#else /* G_BIG_ENDIAN */
+#define BYTE_1_OFFSET 24
+#define BYTE_2_OFFSET 16
+#define BYTE_3_OFFSET 8
+#define BYTE_4_OFFSET 0
+#endif
+
+/* These are the offsets to shift to access the alpha, red, green and blue bytes
+ * in a native endian ARGB value (such as what is given by Cairo) */
+#define ARGB_A_OFFSET 24
+#define ARGB_R_OFFSET 16
+#define ARGB_G_OFFSET 8
+#define ARGB_B_OFFSET 0
+
+/* These are the A,R,G,B component byte masks for a native endian ARGB value
+ * (such as what is given by Cairo) */
+#define ARGB_A_MASK 0xFF << ARGB_A_OFFSET
+#define ARGB_R_MASK 0xFF << ARGB_R_OFFSET
+#define ARGB_G_MASK 0xFF << ARGB_G_OFFSET
+#define ARGB_B_MASK 0xFF << ARGB_B_OFFSET
+
+GST_DEBUG_CATEGORY_STATIC (pgm_gl_text_debug);
+#define GST_CAT_DEFAULT pgm_gl_text_debug
 
 static PgmGlDrawableClass *parent_class = NULL;
 
@@ -90,16 +124,10 @@ set_coordinates (PgmGlText *gltext)
 {
   PgmTexture *texture = gltext->texture;
 
-  GST_OBJECT_LOCK (gltext);
-  PGM_TEXTURE_LOCK (texture);
-
   gltext->coord[2] = (PgmGlFloat) texture->width / texture->width_pot;
   gltext->coord[4] = gltext->coord[2];
   gltext->coord[5] = (PgmGlFloat) texture->height / texture->height_pot;
   gltext->coord[7] = gltext->coord[5];
-
-  PGM_TEXTURE_UNLOCK (texture);
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Stores outline width depending on PgmText outline width converting it
@@ -109,17 +137,12 @@ set_outline_width (PgmGlText *gltext)
 {
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
-  gfloat outline_width;
 
-  outline_width = get_projected_height (gldrawable, text->outline_width);
-
-  GST_OBJECT_LOCK (gltext);
-  gltext->outline_width = outline_width;
-  GST_OBJECT_UNLOCK (gltext);
+  gltext->outline_width = get_projected_height (gldrawable, text->outline_width);
 }
 
 /* Stores outline color depending on PgmText color normalizing components to
- * the range [0.0; 0.1] */
+ * the range [0.0; 1.0] */
 static void
 set_outline_color (PgmGlText *gltext)
 {
@@ -127,14 +150,10 @@ set_outline_color (PgmGlText *gltext)
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
   GST_OBJECT_LOCK (text);
-  GST_OBJECT_LOCK (gltext);
-
   gltext->outline_color[0] = text->outline_r * INV_255;
   gltext->outline_color[1] = text->outline_g * INV_255;
   gltext->outline_color[2] = text->outline_b * INV_255;
   gltext->outline_color[3] = text->outline_a * INV_255;
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (text);
 }
 
@@ -148,29 +167,72 @@ set_line_spacing (PgmGlText *gltext)
   gfloat line_spacing;
 
   line_spacing = get_projected_height (gldrawable, text->line_spacing);
+  pango_layout_set_spacing (gltext->layout, (gint)
+                            (line_spacing * PANGO_SCALE));
+}
 
-  GST_OBJECT_LOCK (gltext);
-  pango_layout_set_spacing (gltext->layout, line_spacing * PANGO_SCALE);
-  GST_OBJECT_UNLOCK (gltext);
+/* Calculates the maximum total height of a line of text for the current font
+ * and font height set in gltext. Returns the value in device unit. */
+static gfloat
+get_max_height (PgmGlText *gltext)
+{
+  /* FIXME: we use the language of the current locale, but the text we display
+   *        may be in another language. */
+  PangoLanguage *language = NULL;
+  PangoFontMetrics *metrics = NULL;
+  gint ascent, descent;
+  gchar *locale = NULL;
+
+#ifdef HAVE_PANGO_1_16
+  language = pango_language_get_default ();
+#else
+#ifdef G_OS_WIN32
+  locale = g_win32_getlocale ();
+#else
+  locale = g_strdup (setlocale (LC_CTYPE, NULL));
+#endif
+  language = pango_language_from_string (locale);
+#endif
+
+  metrics = pango_context_get_metrics (gltext->pango_ctx,
+                                       gltext->desc, language);
+  ascent = pango_font_metrics_get_ascent (metrics);
+  descent = pango_font_metrics_get_descent (metrics);
+
+  pango_font_metrics_unref (metrics);
+
+  g_free (locale);
+
+  return ((ascent + descent) / ((gfloat) PANGO_SCALE));
 }
 
 /* Defines Pango font size depending on PgmText height converting it from
- * canvas coordinates to viewport coordinates */
+ * canvas coordinates to viewport coordinates. This function depends on
+ * various parameters, including the font set for gltext. When setting many
+ * parameters, it is cautious to set this one last. */
 static void
 set_font_height (PgmGlText *gltext)
 {
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
-  gfloat height;
 
-  height = get_projected_height (gldrawable, text->height);
+  /* All the heights here are in device unit. */
+  gfloat requested_height, wanted_height, max_height, font_ratio;
 
-  GST_OBJECT_LOCK (gltext);
+  requested_height = get_projected_height (gldrawable, text->height);
+  requested_height *= gldrawable->height;
 
-  pango_font_description_set_absolute_size (gltext->desc, height * PANGO_SCALE);
+  /* The "pango height" is something like the height of an M, but we want to set
+   * the maximum height of a line, so we need to calculate the ratio first. */
+  pango_font_description_set_absolute_size (gltext->desc,
+                                            requested_height * PANGO_SCALE);
+  max_height = get_max_height (gltext);
+  font_ratio = requested_height / max_height;
+
+  wanted_height = requested_height * font_ratio;
+  pango_font_description_set_absolute_size (gltext->desc,
+                                            wanted_height * PANGO_SCALE);
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango font family depending on PgmText font family */
@@ -181,12 +243,8 @@ set_font_family (PgmGlText *gltext)
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
   GST_OBJECT_LOCK (text);
-  GST_OBJECT_LOCK (gltext);
-
   pango_font_description_set_family (gltext->desc, text->font_family);
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (text);
 }
 
@@ -197,17 +255,13 @@ set_label (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (text);
-  GST_OBJECT_LOCK (gltext);
-
   /* FIXME: defining a markup and right after a label keeps the
    *        previous markups on the new label. */
+  GST_OBJECT_LOCK (text);
   if (text->use_markup)
     pango_layout_set_markup (gltext->layout, text->label, -1);
   else
     pango_layout_set_text (gltext->layout, text->label, -1);
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (text);
 }
 
@@ -222,16 +276,12 @@ set_width (PgmGlText *gltext)
 
   width = get_projected_width (gldrawable, gldrawable->width);
 
-  GST_OBJECT_LOCK (gltext);
-
   GST_OBJECT_LOCK (viewport);
   non_square_ratio = 1.0f / viewport->pixel_aspect_ratio;
   GST_OBJECT_UNLOCK (viewport);
 
-  pango_layout_set_width (gltext->layout,
-                          width * PANGO_SCALE * non_square_ratio);
-
-  GST_OBJECT_UNLOCK (gltext);
+  pango_layout_set_width (gltext->layout, (guint)
+                          (width * PANGO_SCALE * non_square_ratio));
 }
 
 /* Defines Pango justification depending on PgmText justification */
@@ -243,11 +293,7 @@ set_justify (PgmGlText *gltext)
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
   GST_OBJECT_LOCK (text);
-  GST_OBJECT_LOCK (gltext);
-
   pango_layout_set_justify (gltext->layout, text->justify);
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (text);
 #endif /* HAVE_PANGO_1_16 */
 }
@@ -259,8 +305,6 @@ set_weight (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->weight)
     {
     case PGM_TEXT_WEIGHT_LIGHT:
@@ -280,8 +324,6 @@ set_weight (PgmGlText *gltext)
     }
 
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango stretching depending on PgmText stretching */
@@ -291,8 +333,6 @@ set_stretch (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->stretch)
     {
     case PGM_TEXT_STRETCH_CONDENSED:
@@ -314,8 +354,6 @@ set_stretch (PgmGlText *gltext)
     }
 
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango alignment depending on PgmText alignment */
@@ -325,8 +363,6 @@ set_alignment (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->alignment)
     {
     case PGM_TEXT_ALIGN_LEFT:
@@ -344,8 +380,6 @@ set_alignment (PgmGlText *gltext)
     default:
       break;
     }
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango ellipsizing depending on PgmText ellipsizing */
@@ -355,8 +389,6 @@ set_ellipsize (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->ellipsize)
     {
     case PGM_TEXT_ELLIPSIZE_NONE:
@@ -378,8 +410,6 @@ set_ellipsize (PgmGlText *gltext)
     default:
       break;
     }
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango wrapping depending on PgmText wrapping */
@@ -389,8 +419,6 @@ set_wrap (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->wrap)
     {
     case PGM_TEXT_WRAP_WORD:
@@ -408,8 +436,6 @@ set_wrap (PgmGlText *gltext)
     default:
       break;
     }
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango gravity depending on PgmText gravity */
@@ -420,8 +446,6 @@ set_gravity (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->gravity)
     {
     case PGM_TEXT_GRAVITY_SOUTH:
@@ -447,8 +471,6 @@ set_gravity (PgmGlText *gltext)
     default:
       break;
     }
-
-  GST_OBJECT_UNLOCK (gltext);
 #endif /* HAVE_PANGO_1_16 */
 }
 
@@ -459,8 +481,6 @@ set_style (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->style)
     {
     case PGM_TEXT_STYLE_NORMAL:
@@ -480,8 +500,6 @@ set_style (PgmGlText *gltext)
     }
 
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Defines Pango variant depending on PgmText variant */
@@ -491,8 +509,6 @@ set_variant (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmText *text = PGM_TEXT (gldrawable->drawable);
 
-  GST_OBJECT_LOCK (gltext);
-
   switch (text->variant)
     {
     case PGM_TEXT_VARIANT_NORMAL:
@@ -509,16 +525,14 @@ set_variant (PgmGlText *gltext)
     }
 
   pango_layout_set_font_description (gltext->layout, gltext->desc);
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 /* Update all the properties of the text */
 static void
 update_properties (PgmGlText *gltext)
 {
-  set_font_height (gltext);
   set_font_family (gltext);
+  set_weight (gltext);
   set_stretch (gltext);
   set_style (gltext);
   set_variant (gltext);
@@ -532,10 +546,12 @@ update_properties (PgmGlText *gltext)
   set_label (gltext);
   set_outline_width (gltext);
   set_outline_color (gltext);
+  set_font_height (gltext);
 }
 
 /* Cairo uses premultiplied-alpha to represent pixels (mostly for compositing).
- * This function unpremultiplies the alpha of the native endian ARGB buffer. */
+ * This function unpremultiplies the alpha of cairo's native endian ARGB buffer
+ * and converts it to BGRA in memory order. */
 static void
 unpremultiply_alpha (guint32 *buffer,
                      guint width,
@@ -547,13 +563,32 @@ unpremultiply_alpha (guint32 *buffer,
 
   for (i = 0; i < size; i++)
     {
-      a = (buffer[i] & 0xff000000) >> 24;
-      if (a > 0 && a < 255)
+      a = (buffer[i] & ARGB_A_MASK) >> ARGB_A_OFFSET;
+      if (a == 0) /* optimisation for most common case */
         {
-          r = (((buffer[i] & 0x00ff0000) >> 16) * 255 + a / 2) / a;
-          g = (((buffer[i] & 0x0000ff00) >>  8) * 255 + a / 2) / a;
-          b = (((buffer[i] & 0x000000ff) >>  0) * 255 + a / 2) / a;
-          buffer[i] = a << 24 | r << 16 | g << 8 | b;
+          continue;
+        }
+      else if (a == 255)
+        {
+          /* If the pixel is full alpha, we don't have to change the rgb values,
+           * but we have to convert the pixel from native endian ARGB to memory
+           * BGRA, which is equivalent to little endian ARGB. */
+          buffer[i] = GUINT32_TO_LE (buffer[i]);
+        }
+      else
+        {
+          r = (((buffer[i] & ARGB_R_MASK) >> ARGB_R_OFFSET)
+               * 255 + a / 2) / a;
+          g = (((buffer[i] & ARGB_G_MASK) >> ARGB_G_OFFSET)
+               * 255 + a / 2) / a;
+          b = (((buffer[i] & ARGB_B_MASK) >> ARGB_B_OFFSET)
+               * 255 + a / 2) / a;
+          /* we write the buffer in BGRA memory order (B in low address, A in
+           * high address) in one memory access */
+          buffer[i] = b << BYTE_1_OFFSET
+              | g << BYTE_2_OFFSET
+              | r << BYTE_3_OFFSET
+              | a << BYTE_4_OFFSET;
         }
     }
 }
@@ -568,10 +603,10 @@ create_pixmap (PgmGlText *gltext)
   PgmContextTask *task;
   gfloat pixel_aspect_ratio;
 
-  gltext->width = get_projected_width (gldrawable, gldrawable->width);
-  gltext->height = get_projected_height (gldrawable, gldrawable->height);
-
-  GST_OBJECT_LOCK (gltext);
+  gltext->width = (guint) get_projected_width (gldrawable,
+                                               gldrawable->width);
+  gltext->height = (guint) get_projected_height (gldrawable,
+                                                 gldrawable->height);
 
   GST_OBJECT_LOCK (viewport);
   pixel_aspect_ratio = viewport->pixel_aspect_ratio;
@@ -586,15 +621,13 @@ create_pixmap (PgmGlText *gltext)
                                                          gltext->width * 4);
 
   gltext->cairo_ctx = cairo_create (gltext->surface);
+  cairo_set_font_options (gltext->cairo_ctx, gltext->font_options);
   cairo_scale (gltext->cairo_ctx, pixel_aspect_ratio, 1.0f);
   gltext->layout = pango_cairo_create_layout (gltext->cairo_ctx);
   gltext->pango_ctx = pango_layout_get_context (gltext->layout);
 
   pgm_texture_set_buffer (gltext->texture, gltext->buffer, PGM_IMAGE_BGRA,
                           gltext->width, gltext->height, gltext->size, 0);
-
-  GST_OBJECT_UNLOCK (gltext);
-
   set_coordinates (gltext);
 
   /* And request a texture generation */
@@ -609,8 +642,6 @@ free_pixmap (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmContextTask *task;
 
-  GST_OBJECT_LOCK (gltext);
-
   g_object_unref (gltext->layout);
   gltext->layout = NULL;
   cairo_destroy (gltext->cairo_ctx);
@@ -623,8 +654,6 @@ free_pixmap (PgmGlText *gltext)
       gltext->buffer = NULL;
     }
 
-  GST_OBJECT_UNLOCK (gltext);
-
   /* Request texture clean up */
   task = pgm_context_task_new (PGM_CONTEXT_CLEAN_TEXTURE, gltext->texture);
   pgm_context_push_immediate_task (gldrawable->glviewport->context, task);
@@ -637,8 +666,6 @@ update_pixmap (PgmGlText *gltext)
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (gltext);
   PgmContextTask *task;
 
-  GST_OBJECT_LOCK (gltext);
-
   /* Make context fully transparent */
   cairo_set_operator (gltext->cairo_ctx, CAIRO_OPERATOR_SOURCE);
   cairo_set_source_rgba (gltext->cairo_ctx, 0.0f, 0.0f, 0.0f, 0.0f);
@@ -661,11 +688,13 @@ update_pixmap (PgmGlText *gltext)
       cairo_stroke (gltext->cairo_ctx);
     }
 
-  GST_OBJECT_UNLOCK (gltext);
-
 #if 0
   /* Debugging */
+#ifndef WIN32
   cairo_surface_write_to_png (gltext->surface, "/tmp/pgmgltext.png");
+#else
+  cairo_surface_write_to_png (gltext->surface, "C:\\pgmgltext.png");
+#endif /* WIN32 */
 #endif
 
   /* Get rid of the premultiplied alpha */
@@ -692,20 +721,11 @@ regenerate_text (PgmGlText *gltext)
 static void
 update_text (PgmGlText *gltext)
 {
-  gboolean need_regeneration;
-
-  GST_OBJECT_LOCK (gltext);
-  need_regeneration = gltext->size_updated;
-  GST_OBJECT_UNLOCK (gltext);
-
-  if (!need_regeneration)
+  if (!gltext->size_updated)
     update_pixmap (gltext);
   else
     {
-      GST_OBJECT_LOCK (gltext);
       gltext->size_updated = FALSE;
-      GST_OBJECT_UNLOCK (gltext);
-
       regenerate_text (gltext);
     }
 }
@@ -716,49 +736,38 @@ static void
 pgm_gl_text_draw (PgmGlDrawable *gldrawable)
 {
   PgmGlText *gltext = PGM_GL_TEXT (gldrawable);
-  PgmContextProcAddress *gl;
 
-  GST_DEBUG_OBJECT (gldrawable, "draw");
+  GST_LOG_OBJECT (gldrawable, "draw");
 
-  gl = gldrawable->glviewport->context->gl;
+  /* Only draw if it's visible */
+  if (gltext->fg_color[3] != 0.0f)
+    {
+      PgmContextProcAddress *gl = gldrawable->glviewport->context->gl;
 
-  pgm_texture_bind (gltext->texture);
-
-  gl->enable_client_state (PGM_GL_VERTEX_ARRAY);
-  gl->enable_client_state (PGM_GL_TEXTURE_COORD_ARRAY);
-
-  GST_OBJECT_LOCK (gltext);
-  gl->vertex_pointer (3, PGM_GL_FLOAT, 0, gltext->vertex);
-  gl->tex_coord_pointer (2, PGM_GL_FLOAT, 0, gltext->coord);
-  gl->color_4fv (gltext->fg_color);
-  GST_OBJECT_UNLOCK (gltext);
-
-  gl->draw_arrays (PGM_GL_QUADS, 0, 4);
-  gl->disable_client_state (PGM_GL_VERTEX_ARRAY);
-  gl->disable_client_state (PGM_GL_TEXTURE_COORD_ARRAY);
-
-  pgm_texture_unbind (gltext->texture);
+      pgm_texture_bind (gltext->texture);
+      gl->enable_client_state (PGM_GL_VERTEX_ARRAY);
+      gl->enable_client_state (PGM_GL_TEXTURE_COORD_ARRAY);
+      gl->vertex_pointer (3, PGM_GL_FLOAT, 0, gltext->vertex);
+      gl->tex_coord_pointer (2, PGM_GL_FLOAT, 0, gltext->coord);
+      gl->color_4fv (gltext->fg_color);
+      gl->draw_arrays (PGM_GL_QUADS, 0, 4);
+      gl->disable_client_state (PGM_GL_VERTEX_ARRAY);
+      gl->disable_client_state (PGM_GL_TEXTURE_COORD_ARRAY);
+      pgm_texture_unbind (gltext->texture);
+    }
 }
 
 static void
 pgm_gl_text_regenerate (PgmGlDrawable *gldrawable)
 {
   PgmGlText *gltext = PGM_GL_TEXT (gldrawable);
-  gboolean updated;
 
   GST_DEBUG_OBJECT (gldrawable, "regenerate");
 
-  GST_OBJECT_LOCK (gltext);
-  updated = gltext->size_updated;
-  GST_OBJECT_UNLOCK (gltext);
-
-  if (updated)
+  if (gltext->size_updated)
     {
       regenerate_text (gltext);
-
-      GST_OBJECT_LOCK (gltext);
       gltext->size_updated = FALSE;
-      GST_OBJECT_UNLOCK (gltext);
     }
 }
 
@@ -769,9 +778,7 @@ pgm_gl_text_update_projection (PgmGlDrawable *gldrawable)
 
   GST_DEBUG_OBJECT (gldrawable, "update_projection");
 
-  GST_OBJECT_LOCK (gltext);
   gltext->size_updated = TRUE;
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 static void
@@ -782,14 +789,10 @@ pgm_gl_text_set_size (PgmGlDrawable *gldrawable)
 
   GST_DEBUG_OBJECT (gldrawable, "set_size");
 
-  GST_OBJECT_LOCK (gltext);
-
   for (i = 0; i < 12; i++)
     gltext->vertex[i] = gldrawable->bg_vertex[i];
 
   gltext->size_updated = TRUE;
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 static void
@@ -800,12 +803,8 @@ pgm_gl_text_set_position (PgmGlDrawable *gldrawable)
 
   GST_DEBUG_OBJECT (gldrawable, "set_position");
 
-  GST_OBJECT_LOCK (gltext);
-
   for (i = 0; i < 12; i++)
     gltext->vertex[i] = gldrawable->bg_vertex[i];
-
-  GST_OBJECT_UNLOCK (gltext);
 }
 
 static void
@@ -819,14 +818,10 @@ pgm_gl_text_set_fg_color (PgmGlDrawable *gldrawable)
   drawable = gldrawable->drawable;
 
   GST_OBJECT_LOCK (drawable);
-  GST_OBJECT_LOCK (gltext);
-
   gltext->fg_color[0] = drawable->fg_r * INV_255;
   gltext->fg_color[1] = drawable->fg_g * INV_255;
   gltext->fg_color[2] = drawable->fg_b * INV_255;
   gltext->fg_color[3] = drawable->fg_a * drawable->opacity * SQR_INV_255;
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (drawable);
 }
 
@@ -841,11 +836,7 @@ pgm_gl_text_set_opacity (PgmGlDrawable *gldrawable)
   drawable = gldrawable->drawable;
 
   GST_OBJECT_LOCK (drawable);
-  GST_OBJECT_LOCK (gltext);
-
   gltext->fg_color[3] = drawable->fg_a * drawable->opacity * SQR_INV_255;
-
-  GST_OBJECT_UNLOCK (gltext);
   GST_OBJECT_UNLOCK (drawable);
 }
 
@@ -868,7 +859,13 @@ pgm_gl_text_sync (PgmGlDrawable *gldrawable)
 
 /* GObject stuff */
 
-G_DEFINE_TYPE (PgmGlText, pgm_gl_text, PGM_TYPE_GL_DRAWABLE);
+PGM_DEFINE_DYNAMIC_TYPE (PgmGlText, pgm_gl_text, PGM_TYPE_GL_DRAWABLE);
+
+void
+pgm_gl_text_register (GTypeModule *module)
+{
+  pgm_gl_text_register_type (module);
+}
 
 static void
 pgm_gl_text_dispose (GObject *object)
@@ -880,7 +877,12 @@ pgm_gl_text_dispose (GObject *object)
   GST_DEBUG_OBJECT (gltext, "dispose");
 
   free_pixmap (gltext);
+
   pango_font_description_free (gltext->desc);
+  gltext->desc = NULL;
+
+  cairo_font_options_destroy (gltext->font_options);
+  gltext->font_options = NULL;
 
   gst_object_unref (gldrawable->drawable);
   pgm_context_remove_tasks_with_data (gldrawable->glviewport->context,
@@ -898,6 +900,9 @@ pgm_gl_text_class_init (PgmGlTextClass *klass)
   GObjectClass *gobject_class;
   PgmGlDrawableClass *drawable_class;
 
+  GST_DEBUG_CATEGORY_INIT (pgm_gl_text_debug, "pgm_gl_text", 0,
+                           "OpenGL plugin: PgmGlText");
+
   parent_class = g_type_class_peek_parent (klass);
 
   gobject_class = G_OBJECT_CLASS (klass);
@@ -918,12 +923,23 @@ pgm_gl_text_class_init (PgmGlTextClass *klass)
   drawable_class->set_opacity = GST_DEBUG_FUNCPTR (pgm_gl_text_set_opacity);
 }
 
+static void
+pgm_gl_text_class_finalize (PgmGlTextClass *klass)
+{
+  return;
+}
+
 static void
 pgm_gl_text_init (PgmGlText *gltext)
 {
   GST_DEBUG_OBJECT (gltext, "init");
 
+  gltext->font_options = cairo_font_options_create ();
+  cairo_font_options_set_antialias (gltext->font_options,
+                                    CAIRO_ANTIALIAS_SUBPIXEL);
+
   gltext->desc = pango_font_description_new ();
+
   gltext->size_updated = FALSE;
 }
 
@@ -945,6 +961,7 @@ pgm_gl_text_new (PgmDrawable *drawable,
   gldrawable = PGM_GL_DRAWABLE (gltext);
   gldrawable->drawable = gst_object_ref (drawable);
   gldrawable->glviewport = glviewport;
+  pgm_gl_viewport_connect_changed_callback (glviewport, gldrawable);
   pgm_gl_drawable_sync (gldrawable);
 
   return gldrawable;
