@@ -2,7 +2,7 @@
  *
  * Pigment OpenGL plugin
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -36,7 +36,7 @@
  * GL drawable address as the value. When a "generic" drawable property changes,
  * it emits the 'changed' signal to which the Gl Viewport is connected. The
  * handler stores the change as a task in an update queue which is flushed
- * when the user request a viewport update through pgm_viewport_update().
+ * during the automatic update done in the rendering thread in PgmContext.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -47,8 +47,8 @@
 #include "pgmglimage.h"
 #include "pgmgltext.h"
 
-GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
-#define GST_CAT_DEFAULT pgm_gl_debug
+GST_DEBUG_CATEGORY_STATIC (pgm_gl_viewport_debug);
+#define GST_CAT_DEFAULT pgm_gl_viewport_debug
 
 /* Task types */
 typedef enum {
@@ -305,22 +305,14 @@ gl_drawable_new (PgmGlViewport *glviewport,
   else
     {
       gldrawable = NULL;
-      GST_DEBUG_OBJECT (glviewport, "cannot create object from this type");
+      GST_WARNING_OBJECT (glviewport, "cannot create object from this type");
     }
 
-  /* Make it easily retrievable and connect to the 'changed' signal so that
-   * changes could be reflected in the plugin */
+  /* Make it easily retrievable */
   if (gldrawable)
     {
       GST_OBJECT_LOCK (glviewport);
-      GST_OBJECT_LOCK (drawable);
-
       g_hash_table_insert (glviewport->drawable_hash, drawable, gldrawable);
-      gldrawable->change_handler = g_signal_connect (drawable, "changed",
-                                                     G_CALLBACK (changed_cb),
-                                                     (gpointer) gldrawable);
-
-      GST_OBJECT_UNLOCK (drawable);
       GST_OBJECT_UNLOCK (glviewport);
     }
 
@@ -367,7 +359,7 @@ gl_drawable_free (PgmGlViewport *glviewport,
   g_mutex_unlock (glviewport->update_lock);
 
   /* Then delete it */
-  GST_DEBUG_OBJECT (glviewport, "deleting %s", GST_OBJECT_NAME (gldrawable));
+  GST_DEBUG_OBJECT (glviewport, "unreferencing %s", GST_OBJECT_NAME (gldrawable));
   gst_object_unref (gldrawable);
 }
 
@@ -403,21 +395,27 @@ do_task_reorder (PgmGlViewport *glviewport,
   switch (task->reorder.layer)
     {
     case PGM_DRAWABLE_NEAR:
-      glviewport->near = g_list_remove (glviewport->near, gldrawable);
-      glviewport->near = g_list_insert (glviewport->near, gldrawable,
-                                        task->reorder.order);
+      glviewport->near_layer = g_list_remove (glviewport->near_layer,
+                                              gldrawable);
+      glviewport->near_layer = g_list_insert (glviewport->near_layer,
+                                              gldrawable,
+                                              task->reorder.order);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      glviewport->middle = g_list_remove (glviewport->middle, gldrawable);
-      glviewport->middle = g_list_insert (glviewport->middle, gldrawable,
-                                          task->reorder.order);
+      glviewport->middle_layer = g_list_remove (glviewport->middle_layer,
+                                                gldrawable);
+      glviewport->middle_layer = g_list_insert (glviewport->middle_layer,
+                                                gldrawable,
+                                                task->reorder.order);
       break;
 
     case PGM_DRAWABLE_FAR:
-      glviewport->far = g_list_remove (glviewport->far, gldrawable);
-      glviewport->far = g_list_insert (glviewport->far, gldrawable,
-                                       task->reorder.order);
+      glviewport->far_layer = g_list_remove (glviewport->far_layer,
+                                             gldrawable);
+      glviewport->far_layer = g_list_insert (glviewport->far_layer,
+                                             gldrawable,
+                                             task->reorder.order);
       break;
 
     default:
@@ -443,18 +441,21 @@ do_task_add (PgmGlViewport *glviewport,
   switch (task->add.layer)
     {
     case PGM_DRAWABLE_NEAR:
-      glviewport->near = g_list_insert (glviewport->near, gldrawable,
-                                        task->add.order);
+      glviewport->near_layer = g_list_insert (glviewport->near_layer,
+                                              gldrawable,
+                                              task->add.order);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      glviewport->middle = g_list_insert (glviewport->middle, gldrawable,
-                                          task->add.order);
+      glviewport->middle_layer = g_list_insert (glviewport->middle_layer,
+                                                gldrawable,
+                                                task->add.order);
       break;
 
     case PGM_DRAWABLE_FAR:
-      glviewport->far = g_list_insert (glviewport->far, gldrawable,
-                                       task->add.order);
+      glviewport->far_layer = g_list_insert (glviewport->far_layer,
+                                             gldrawable,
+                                             task->add.order);
       break;
 
     default:
@@ -483,15 +484,18 @@ do_task_remove (PgmGlViewport *glviewport,
   switch (task->remove.layer)
     {
     case PGM_DRAWABLE_NEAR:
-      glviewport->near = g_list_remove (glviewport->near, gldrawable);
+      glviewport->near_layer = g_list_remove (glviewport->near_layer,
+                                              gldrawable);
       break;
 
     case PGM_DRAWABLE_MIDDLE:
-      glviewport->middle = g_list_remove (glviewport->middle, gldrawable);
+      glviewport->middle_layer = g_list_remove (glviewport->middle_layer,
+                                                gldrawable);
       break;
 
     case PGM_DRAWABLE_FAR:
-      glviewport->far = g_list_remove (glviewport->far, gldrawable);
+      glviewport->far_layer = g_list_remove (glviewport->far_layer,
+                                             gldrawable);
       break;
 
     default:
@@ -512,9 +516,12 @@ do_task_regenerate (PgmGlViewport *glviewport,
 {
   g_mutex_lock (glviewport->layer_lock);
 
-  g_list_foreach (glviewport->near, (GFunc) pgm_gl_drawable_regenerate, NULL);
-  g_list_foreach (glviewport->middle, (GFunc) pgm_gl_drawable_regenerate, NULL);
-  g_list_foreach (glviewport->far, (GFunc) pgm_gl_drawable_regenerate, NULL);
+  g_list_foreach (glviewport->near_layer,
+                  (GFunc) pgm_gl_drawable_regenerate, NULL);
+  g_list_foreach (glviewport->middle_layer,
+                  (GFunc) pgm_gl_drawable_regenerate, NULL);
+  g_list_foreach (glviewport->far_layer,
+                  (GFunc) pgm_gl_drawable_regenerate, NULL);
 
   g_mutex_unlock (glviewport->layer_lock);
 
@@ -530,7 +537,7 @@ changed_cb (PgmDrawable *drawable,
   PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (data);
   PgmGlViewport *glviewport = gldrawable->glviewport;
 
-  GST_DEBUG_OBJECT (glviewport, "drawable_changed_cb");
+  GST_LOG_OBJECT (glviewport, "drawable_changed_cb");
 
   if (glviewport)
     {
@@ -542,6 +549,9 @@ changed_cb (PgmDrawable *drawable,
       glviewport->update_queue = g_slist_prepend (glviewport->update_queue,
                                                   task);
       g_mutex_unlock (glviewport->update_lock);
+
+      /* Request an update */
+      pgm_context_update (glviewport->context);
     }
 }
 
@@ -563,6 +573,9 @@ drawable_added_cb (PgmCanvas *canvas,
   g_mutex_lock (glviewport->update_lock);
   glviewport->update_queue = g_slist_prepend (glviewport->update_queue, task);
   g_mutex_unlock (glviewport->update_lock);
+
+  /* Request an update */
+  pgm_context_update (glviewport->context);
 }
 
 /* Canvas 'drawable-removed' handler */
@@ -582,6 +595,9 @@ drawable_removed_cb (PgmCanvas *canvas,
   g_mutex_lock (glviewport->update_lock);
   glviewport->update_queue = g_slist_prepend (glviewport->update_queue, task);
   g_mutex_unlock (glviewport->update_lock);
+
+  /* Request an update */
+  pgm_context_update (glviewport->context);
 }
 
 /* Canvas 'drawable-reordered' handler */
@@ -602,6 +618,9 @@ drawable_reordered_cb (PgmCanvas *canvas,
   g_mutex_lock (glviewport->update_lock);
   glviewport->update_queue = g_slist_prepend (glviewport->update_queue, task);
   g_mutex_unlock (glviewport->update_lock);
+
+  /* Request an update */
+  pgm_context_update (glviewport->context);
 }
 
 /* Canvas 'regenerated' handler */
@@ -619,6 +638,9 @@ regenerated_cb (PgmCanvas *canvas,
   g_mutex_lock (glviewport->update_lock);
   glviewport->update_queue = g_slist_prepend (glviewport->update_queue, task);
   g_mutex_unlock (glviewport->update_lock);
+
+  /* Request an update */
+  pgm_context_update (glviewport->context);
 }
 
 /* Delete all GL drawable from the given GL viewport layer list and free it */
@@ -652,34 +674,37 @@ sync_gl_drawable_from_canvas (PgmGlViewport *glviewport,
   GList *walk;
 
   /* Sync near layer */
-  walk = canvas->near;
+  walk = canvas->near_layer;
   while (walk)
     {
       gldrawable = gl_drawable_new (glviewport, (PgmDrawable*) walk->data);
       g_mutex_lock (glviewport->layer_lock);
-      glviewport->near = g_list_append (glviewport->near, gldrawable);
+      glviewport->near_layer = g_list_append (glviewport->near_layer,
+                                              gldrawable);
       g_mutex_unlock (glviewport->layer_lock);
       walk = walk->next;
     }
 
   /* Sync middle layer */
-  walk = canvas->middle;
+  walk = canvas->middle_layer;
   while (walk)
     {
       gldrawable = gl_drawable_new (glviewport, (PgmDrawable*) walk->data);
       g_mutex_lock (glviewport->layer_lock);
-      glviewport->middle = g_list_append (glviewport->middle, gldrawable);
+      glviewport->middle_layer = g_list_append (glviewport->middle_layer,
+                                                gldrawable);
       g_mutex_unlock (glviewport->layer_lock);
       walk = walk->next;
     }
 
   /* Sync far layer */
-  walk = canvas->far;
+  walk = canvas->far_layer;
   while (walk)
     {
       gldrawable = gl_drawable_new (glviewport, (PgmDrawable*) walk->data);
       g_mutex_lock (glviewport->layer_lock);
-      glviewport->far = g_list_append (glviewport->far, gldrawable);
+      glviewport->far_layer = g_list_append (glviewport->far_layer,
+                                             gldrawable);
       g_mutex_unlock (glviewport->layer_lock);
       walk = walk->next;
     }
@@ -687,19 +712,6 @@ sync_gl_drawable_from_canvas (PgmGlViewport *glviewport,
 
 /* PgmViewport methods */
 
-static PgmError
-pgm_gl_viewport_update (PgmViewport *viewport)
-{
-  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
-
-  GST_DEBUG_OBJECT (glviewport, "update");
-
-  /* Request an update in the rendering thread */
-  pgm_context_update (glviewport->context);
-
-  return PGM_ERROR_OK;
-}
-
 static PgmError
 pgm_gl_viewport_set_title (PgmViewport *viewport,
                            const gchar *title)
@@ -719,10 +731,12 @@ static PgmError
 pgm_gl_viewport_show (PgmViewport *viewport)
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
 
   GST_DEBUG_OBJECT (glviewport, "show");
 
-  pgm_backend_show (glviewport->context->backend);
+  task = pgm_context_task_new (PGM_CONTEXT_VISIBILITY, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
 
   return PGM_ERROR_OK;
 }
@@ -731,10 +745,27 @@ static PgmError
 pgm_gl_viewport_hide (PgmViewport *viewport)
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
 
   GST_DEBUG_OBJECT (glviewport, "hide");
 
-  pgm_backend_hide (glviewport->context->backend);
+  task = pgm_context_task_new (PGM_CONTEXT_VISIBILITY, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
+
+  return PGM_ERROR_OK;
+}
+
+static PgmError
+pgm_gl_viewport_set_decorated (PgmViewport *viewport,
+                               gboolean decorated)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glviewport, "set_decorated");
+
+  task = pgm_context_task_new (PGM_CONTEXT_DECORATION, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
 
   return PGM_ERROR_OK;
 }
@@ -769,6 +800,23 @@ pgm_gl_viewport_set_icon (PgmViewport *viewport,
   return PGM_ERROR_OK;
 }
 
+static PgmError
+pgm_gl_viewport_set_drag_status (PgmViewport *viewport,
+                                 gboolean accept)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glviewport, "set_drag_status");
+
+  glviewport->drag_status = accept;
+
+  task = pgm_context_task_new (PGM_CONTEXT_DRAG_STATUS, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
+
+  return PGM_ERROR_OK;
+}
+
 static PgmError
 pgm_gl_viewport_set_size (PgmViewport *viewport,
                           gint width,
@@ -785,6 +833,36 @@ pgm_gl_viewport_set_size (PgmViewport *viewport,
   return PGM_ERROR_OK;
 }
 
+static PgmError
+pgm_gl_viewport_set_alpha_blending (PgmViewport *viewport,
+                                    gboolean alpha_blending)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glviewport, "set_alpha_blending");
+
+  task = pgm_context_task_new (PGM_CONTEXT_ALPHA_BLENDING, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
+
+  return PGM_ERROR_OK;
+}
+
+static PgmError
+pgm_gl_viewport_set_opacity (PgmViewport *viewport,
+                             guchar opacity)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glviewport, "set_opacity");
+
+  task = pgm_context_task_new (PGM_CONTEXT_OPACITY, NULL);
+  pgm_context_push_immediate_task (glviewport->context, task);
+
+  return PGM_ERROR_OK;
+}
+
 static PgmError
 pgm_gl_viewport_set_fullscreen (PgmViewport *viewport,
                                 gboolean fullscreen)
@@ -807,7 +885,7 @@ pgm_gl_viewport_get_screen_resolution (PgmViewport *viewport,
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
 
-  GST_DEBUG_OBJECT (glviewport, "get_screen_resolution");
+  GST_LOG_OBJECT (glviewport, "get_screen_resolution");
 
   pgm_backend_get_screen_resolution (glviewport->context->backend, width,
                                      height);
@@ -822,7 +900,7 @@ pgm_gl_viewport_get_screen_size_mm (PgmViewport *viewport,
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
 
-  GST_DEBUG_OBJECT (glviewport, "get_screen_size_mm");
+  GST_LOG_OBJECT (glviewport, "get_screen_size_mm");
 
   pgm_backend_get_screen_size_mm (glviewport->context->backend, width, height);
 
@@ -857,9 +935,9 @@ pgm_gl_viewport_set_canvas (PgmViewport *viewport,
       GST_OBJECT_UNLOCK (glviewport);
 
       /* Delete our hierarchy */
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->near);
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->middle);
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->far);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->near_layer);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->middle_layer);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->far_layer);
 
       GST_OBJECT_LOCK (glviewport);
       glviewport->canvas = NULL;
@@ -915,14 +993,29 @@ pgm_gl_viewport_update_projection (PgmViewport *viewport)
   return PGM_ERROR_OK;
 }
 
+static PgmError
+pgm_gl_viewport_get_embedding_id (PgmViewport *viewport,
+                                  gulong *embedding_id)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+
+  GST_LOG_OBJECT (glviewport, "get_embedding_id");
+
+  GST_OBJECT_LOCK (glviewport);
+  pgm_backend_get_embedding_id (glviewport->context->backend, embedding_id);
+  GST_OBJECT_UNLOCK (glviewport);
+
+  return PGM_ERROR_OK;
+}
+
 static PgmError
 pgm_gl_viewport_get_pixel_formats (PgmViewport *viewport,
                                    gulong *formats_mask)
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
-  gchar *variable;
+  const gchar *variable;
 
-  GST_DEBUG_OBJECT (glviewport, "get_pixel_formats");
+  GST_LOG_OBJECT (glviewport, "get_pixel_formats");
 
   GST_OBJECT_LOCK (glviewport);
 
@@ -940,7 +1033,7 @@ pgm_gl_viewport_get_pixel_formats (PgmViewport *viewport,
     {
       *formats_mask |= PGM_IMAGE_I420 | PGM_IMAGE_YV12;
 
-      variable = getenv ("PGM_GL_CSP_PROGRAM");
+      variable = g_getenv ("PGM_GL_CSP_PROGRAM");
       if (variable && variable[0] == '0')
         *formats_mask &= ~(PGM_IMAGE_I420 | PGM_IMAGE_YV12);
     }
@@ -958,25 +1051,69 @@ pgm_gl_viewport_get_caps_mask (PgmViewport *viewport,
 {
   PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
 
-  GST_DEBUG_OBJECT (glviewport, "get_caps_mask");
+  GST_LOG_OBJECT (glviewport, "get_caps_mask");
 
   GST_OBJECT_LOCK (glviewport);
-
-  /* Init mask */
-  *caps_mask = 0x0;
-
-  /* And set the acceleration flag */
-  if (pgm_backend_is_accelerated (glviewport->context->backend))
-    *caps_mask |= PGM_VIEWPORT_HARDWARE_ACCELERATION;
-
+  *caps_mask = glviewport->capacities;
   GST_OBJECT_UNLOCK (glviewport);
 
   return PGM_ERROR_OK;
 }
 
+static PgmError
+pgm_gl_viewport_get_frame_rate (PgmViewport *viewport,
+                                guint *frame_rate)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+
+  GST_LOG_OBJECT (glviewport, "get_frame_rate");
+
+  GST_OBJECT_LOCK (glviewport);
+  *frame_rate = glviewport->context->fps;
+  GST_OBJECT_UNLOCK (glviewport);
+
+  return PGM_ERROR_OK;
+}
+
+static PgmError
+pgm_gl_viewport_read_pixels (PgmViewport *viewport,
+                             guint x,
+                             guint y,
+                             guint width,
+                             guint height,
+                             guint8 *pixels)
+{
+  PgmGlViewport *glviewport = PGM_GL_VIEWPORT (viewport);
+  PgmGlViewportPixelRectangle *rectangle = NULL;
+  PgmContextTask *task = NULL;
+
+  GST_DEBUG_OBJECT (glviewport, "read_pixels");
+
+  rectangle = g_slice_new (PgmGlViewportPixelRectangle);
+  if (!rectangle)
+    return PGM_ERROR_X;
+
+  rectangle->x = x;
+  rectangle->y = y;
+  rectangle->width = width;
+  rectangle->height = height;
+  rectangle->pixels = pixels;
+
+  task = pgm_context_task_new (PGM_CONTEXT_READ_PIXELS, rectangle);
+  pgm_context_push_immediate_task (glviewport->context, task);
+
+  return PGM_ERROR_OK;
+}
+
 /* GObject stuff */
 
-G_DEFINE_TYPE (PgmGlViewport, pgm_gl_viewport, PGM_TYPE_VIEWPORT);
+PGM_DEFINE_DYNAMIC_TYPE (PgmGlViewport, pgm_gl_viewport, PGM_TYPE_VIEWPORT);
+
+void
+pgm_gl_viewport_register (GTypeModule *module)
+{
+  pgm_gl_viewport_register_type (module);
+}
 
 static void
 pgm_gl_viewport_dispose (GObject *object)
@@ -985,9 +1122,6 @@ pgm_gl_viewport_dispose (GObject *object)
 
   GST_DEBUG_OBJECT (glviewport, "dispose");
 
-  g_slist_foreach (glviewport->update_queue, (GFunc) task_free, NULL);
-  g_slist_free (glviewport->update_queue);
-
   if (glviewport->canvas)
     {
       /* Disconnect handlers */
@@ -1001,11 +1135,19 @@ pgm_gl_viewport_dispose (GObject *object)
                                    glviewport->regenerated_handler);
 
       /* Delete our hierarchy */
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->near);
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->middle);
-      delete_all_gl_drawable_from_layer (glviewport, &glviewport->far);
+      g_mutex_lock (glviewport->layer_lock);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->near_layer);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->middle_layer);
+      delete_all_gl_drawable_from_layer (glviewport, &glviewport->far_layer);
+      g_mutex_unlock (glviewport->layer_lock);
     }
 
+  g_mutex_lock (glviewport->update_lock);
+  g_slist_foreach (glviewport->update_queue, (GFunc) task_free, NULL);
+  g_slist_free (glviewport->update_queue);
+  glviewport->update_queue = NULL;
+  g_mutex_unlock (glviewport->update_lock);
+
   pgm_context_free (glviewport->context);
 
   g_mutex_free (glviewport->layer_lock);
@@ -1020,6 +1162,9 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
   GObjectClass *gobject_class;
   PgmViewportClass *viewport_class;
 
+  GST_DEBUG_CATEGORY_INIT (pgm_gl_viewport_debug, "pgm_gl_viewport", 0,
+                           "OpenGL plugin: PgmGlViewport");
+
   parent_class = g_type_class_peek_parent (klass);
 
   gobject_class = G_OBJECT_CLASS (klass);
@@ -1029,13 +1174,20 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
   gobject_class->dispose = GST_DEBUG_FUNCPTR (pgm_gl_viewport_dispose);
 
   /* PgmViewport virtual table */
-  viewport_class->update = GST_DEBUG_FUNCPTR (pgm_gl_viewport_update);
   viewport_class->set_title = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_title);
   viewport_class->show = GST_DEBUG_FUNCPTR (pgm_gl_viewport_show);
   viewport_class->hide = GST_DEBUG_FUNCPTR (pgm_gl_viewport_hide);
+  viewport_class->set_decorated =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_decorated);
   viewport_class->set_cursor = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_cursor);
   viewport_class->set_icon = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_icon);
+  viewport_class->set_drag_status =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_drag_status);
   viewport_class->set_size = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_size);
+  viewport_class->set_alpha_blending =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_alpha_blending);
+  viewport_class->set_opacity =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_opacity);
   viewport_class->set_fullscreen =
     GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_fullscreen);
   viewport_class->get_screen_resolution =
@@ -1045,10 +1197,15 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
   viewport_class->set_canvas = GST_DEBUG_FUNCPTR (pgm_gl_viewport_set_canvas);
   viewport_class->update_projection =
     GST_DEBUG_FUNCPTR (pgm_gl_viewport_update_projection);
+  viewport_class->get_embedding_id =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_embedding_id);
   viewport_class->get_pixel_formats =
     GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_pixel_formats);
   viewport_class->get_caps_mask =
     GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_caps_mask);
+  viewport_class->get_frame_rate =
+    GST_DEBUG_FUNCPTR (pgm_gl_viewport_get_frame_rate);
+  viewport_class->read_pixels = GST_DEBUG_FUNCPTR (pgm_gl_viewport_read_pixels);
 
   /* PgmGlDrawable changed table */
   klass->changed_func[PGM_DRAWABLE_VISIBILITY] =
@@ -1063,6 +1220,8 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
     GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_fg_color);
   klass->changed_func[PGM_DRAWABLE_OPACITY] =
     GST_DEBUG_FUNCPTR (pgm_gl_drawable_set_opacity);
+  klass->changed_func[PGM_DRAWABLE_REGENERATE] =
+    GST_DEBUG_FUNCPTR (pgm_gl_drawable_regenerate);
 
   /* PgmGlImage changed table */
   klass->changed_func[PGM_IMAGE_DATA_EMPTY] =
@@ -1071,10 +1230,12 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_buffer);
   klass->changed_func[PGM_IMAGE_DATA_GST_BUFFER] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_gst_buffer);
-  klass->changed_func[PGM_IMAGE_DATA_FD] =
-    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_fd);
+  klass->changed_func[PGM_IMAGE_DATA_PIXBUF] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_pixbuf);
   klass->changed_func[PGM_IMAGE_DATA_IMAGE] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_image);
+  klass->changed_func[PGM_IMAGE_MAPPING_MATRIX] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_mapping_matrix);
   klass->changed_func[PGM_IMAGE_ALIGNMENT] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_alignment);
   klass->changed_func[PGM_IMAGE_LAYOUT] =
@@ -1083,6 +1244,12 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_interp);
   klass->changed_func[PGM_IMAGE_ASPECT_RATIO] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_aspect_ratio);
+  klass->changed_func[PGM_IMAGE_BORDER_WIDTH] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_border_width);
+  klass->changed_func[PGM_IMAGE_BORDER_INNER_COLOR] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_border_inner_color);
+  klass->changed_func[PGM_IMAGE_BORDER_OUTER_COLOR] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_border_outer_color);
 
   /* PgmGlText changed table */
   klass->changed_func[PGM_TEXT_LABEL] =
@@ -1119,6 +1286,12 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_text_set_outline_width);
 }
 
+static void
+pgm_gl_viewport_class_finalize (PgmGlViewportClass *klass)
+{
+  return;
+}
+
 static void
 pgm_gl_viewport_init (PgmGlViewport *glviewport)
 {
@@ -1128,9 +1301,9 @@ pgm_gl_viewport_init (PgmGlViewport *glviewport)
   glviewport->canvas = NULL;
 
   /* Empty lists */
-  glviewport->far = NULL;
-  glviewport->middle = NULL;
-  glviewport->near = NULL;
+  glviewport->far_layer = NULL;
+  glviewport->middle_layer = NULL;
+  glviewport->near_layer = NULL;
   glviewport->layer_lock = g_mutex_new ();
 
   /* Update task queue */
@@ -1142,6 +1315,18 @@ pgm_gl_viewport_init (PgmGlViewport *glviewport)
 
   /* Create context */
   glviewport->context = pgm_context_new (glviewport);
+
+  /* Fill capacities */
+  glviewport->capacities = 0;
+  if (pgm_backend_is_accelerated (glviewport->context->backend))
+    glviewport->capacities |= PGM_VIEWPORT_HARDWARE_ACCELERATION;
+  if (pgm_backend_is_embeddable (glviewport->context->backend))
+    glviewport->capacities |= PGM_VIEWPORT_APPLICATION_EMBEDDING;
+  if (pgm_backend_has_alpha_component (glviewport->context->backend))
+    glviewport->capacities |= PGM_VIEWPORT_OPACITY;
+
+  /* Drag status */
+  glviewport->drag_status = FALSE;
 }
 
 /* Public methods */
@@ -1188,12 +1373,27 @@ pgm_gl_viewport_update_drawable_projection (PgmGlViewport *glviewport)
   g_mutex_lock (glviewport->layer_lock);
 
   /* Adapt position and size of all the drawables */
-  g_list_foreach (glviewport->near,
+  g_list_foreach (glviewport->near_layer,
                   (GFunc) pgm_gl_drawable_update_projection, NULL);
-  g_list_foreach (glviewport->middle,
+  g_list_foreach (glviewport->middle_layer,
                   (GFunc) pgm_gl_drawable_update_projection, NULL);
-  g_list_foreach (glviewport->far,
+  g_list_foreach (glviewport->far_layer,
                   (GFunc) pgm_gl_drawable_update_projection, NULL);
 
   g_mutex_unlock (glviewport->layer_lock);
 }
+
+/* Connects the GL drawable to the "changed" signal. Should be called at the
+ * creation of a new GL drawable to be notified of any change. */
+void
+pgm_gl_viewport_connect_changed_callback (PgmGlViewport *glviewport,
+                                          PgmGlDrawable *gldrawable)
+{
+  PgmDrawable *drawable = gldrawable->drawable;
+
+  GST_OBJECT_LOCK (drawable);
+  gldrawable->change_handler = g_signal_connect (drawable, "changed",
+                                                 G_CALLBACK (changed_cb),
+                                                 (gpointer) gldrawable);
+  GST_OBJECT_UNLOCK (drawable);
+}
