@@ -2,7 +2,7 @@
  *
  * Pigment OpenGL plugin
  *
- * Copyright © 2006, 2007 Fluendo Embedded S.L.
+ * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -41,8 +41,12 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
+#include <string.h>     /* memset */
 #include "pgmtexture.h"
 
+GST_DEBUG_CATEGORY_EXTERN (pgm_gl_debug);
+#define GST_CAT_DEFAULT pgm_gl_debug
+
 /* Texture upload function prototypes */
 static void do_rgb_upload       (PgmTexture *texture, void *buffer);
 static void do_bgr_upload       (PgmTexture *texture, void *buffer);
@@ -87,7 +91,7 @@ static const PgmContext *context = NULL;
 static guint
 get_upper_power_of_two (guint number)
 {
-  gint power_of_two = 1;
+  guint power_of_two = 1;
 
   while (power_of_two < number)
     power_of_two <<= 1;
@@ -95,7 +99,7 @@ get_upper_power_of_two (guint number)
   return power_of_two;
 }
 
-/* Specifies texture parameters of the currently binded texture */
+/* Specifies texture parameters of the currently bound texture */
 static void
 set_texture_parameters (PgmTexture *texture)
 {
@@ -107,48 +111,68 @@ set_texture_parameters (PgmTexture *texture)
                        texture->filter);
 }
 
-/* 24 bits RGB texture creation */
+/* 24 bits RGB texture creation
+ * Stored as RGBA so that the part of the texture not used (pot_size - size)
+ * is opaque. */
 static void
 do_rgb_create (PgmTexture *texture)
 {
+  guint8 *pixels = g_malloc0 (texture->width_pot * texture->height_pot * 4);
+
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
-                    PGM_GL_RGB, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_RGBA, PGM_GL_UNSIGNED_BYTE, pixels);
   set_texture_parameters (texture);
+
+  g_free (pixels);
 }
 
-/* 24 bits BGR texture creation */
+/* 24 bits BGR texture creation.
+ * Stored as BGRA so that the part of the texture not used (pot_size - size)
+ * is opaque. */
 static void
 do_bgr_create (PgmTexture *texture)
 {
+  guint8 *pixels = g_malloc0 (texture->width_pot * texture->height_pot * 4);
+
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
-                    PGM_GL_BGR, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_BGRA, PGM_GL_UNSIGNED_BYTE, pixels);
   set_texture_parameters (texture);
+
+  g_free (pixels);
 }
 
 /* 32 bits RGBA texture creation */
 static void
 do_rgba_create (PgmTexture *texture)
 {
+  guint8 *pixels = g_malloc0 (texture->width_pot * texture->height_pot * 4);
+
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
-                    PGM_GL_RGBA, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_RGBA, PGM_GL_UNSIGNED_BYTE, pixels);
   set_texture_parameters (texture);
+
+  g_free (pixels);
 }
 
 /* 32 bits BGRA texture creation */
 static void
 do_bgra_create (PgmTexture *texture)
 {
+  guint8 *pixels = g_malloc0 (texture->width_pot * texture->height_pot * 4);
+
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
-                    PGM_GL_BGRA, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_BGRA, PGM_GL_UNSIGNED_BYTE, pixels);
   set_texture_parameters (texture);
+
+  g_free (pixels);
 }
 
 /* 3 * 8 bits LUMINANCE textures creation to store 12 bits planar YCbCr FourCC
@@ -159,6 +183,8 @@ do_planar_12_create (PgmTexture *texture)
 {
   guint lower_width_pot = texture->width_pot >> 1;
   guint lower_height_pot = texture->height_pot >> 1;
+  guint8 *y_pixels, *cbcr_pixels;
+  guint size;
 
   /* Get the correct strides and Cb/Cr plane offsets */
   texture->cbcr_height = texture->height / 2;
@@ -168,26 +194,38 @@ do_planar_12_create (PgmTexture *texture)
   texture->cr_offset = texture->cb_offset +
     texture->cbcr_stride * GST_ROUND_UP_2 (texture->height) / 2;
 
-  /* Y plane */
+  /* Initialization pixels area, the area is initialized with black pixels in
+   * the YCbCr color space: (0,0,0)RGB == (16,128,128)YCbCr. We replace in our
+   * specific case the Y component which should be 16 by 0. */
+  y_pixels = g_malloc0 (texture->width_pot * texture->height_pot);
+  size = lower_width_pot * lower_height_pot;
+  cbcr_pixels = g_malloc (size);
+  memset (cbcr_pixels, 128, size);
+
+  /* Y plane creation */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     texture->width_pot, texture->height_pot, 0,
-                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, y_pixels);
   set_texture_parameters (texture);
 
-  /* Cb plane */
+  /* Cb plane creation */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[1]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     lower_width_pot, lower_height_pot, 0,
-                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, cbcr_pixels);
   set_texture_parameters (texture);
 
-  /* Cr plane */
+  /* Cr plane creation */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[2]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     lower_width_pot , lower_height_pot, 0,
-                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, NULL);
+                    PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, cbcr_pixels);
   set_texture_parameters (texture);
+
+  /* Free initialization pixels area */
+  g_free (y_pixels);
+  g_free (cbcr_pixels);
 }
 
 /* FIXME */
@@ -255,13 +293,13 @@ do_planar_12_upload (PgmTexture *texture,
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[1]);
   gl->tex_sub_image_2d (PGM_GL_TEXTURE_2D, 0, 0, 0, texture->cbcr_stride,
                         texture->cbcr_height, PGM_GL_LUMINANCE,
-                        PGM_GL_UNSIGNED_BYTE, buffer + texture->cb_offset);
+                        PGM_GL_UNSIGNED_BYTE, ((guint8*) buffer) + texture->cb_offset);
 
   /* Cr plane */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[2]);
   gl->tex_sub_image_2d (PGM_GL_TEXTURE_2D, 0, 0, 0, texture->cbcr_stride,
                         texture->cbcr_height, PGM_GL_LUMINANCE,
-                        PGM_GL_UNSIGNED_BYTE, buffer + texture->cr_offset);
+                        PGM_GL_UNSIGNED_BYTE, ((guint8*) buffer) + texture->cr_offset);
 }
 
 /* FIXME */
@@ -353,12 +391,20 @@ do_uyvy_unbind (PgmTexture *texture)
 {
 }
 
+/* Update the normalized size fields */
+static void
+update_normalized_size (PgmTexture *texture)
+{
+  texture->norm_width = (gfloat) texture->width / texture->width_pot;
+  texture->norm_height = (gfloat) texture->height / texture->height_pot;
+  texture->inv_norm_width = 1.0f / texture->norm_width;
+  texture->inv_norm_height = 1.0f / texture->norm_height;
+}
+
 /* Frees the current stored buffer if any */
 static void
 free_buffer (PgmTexture *texture)
 {
-  PGM_TEXTURE_LOCK (texture);
-
   switch (texture->storage)
     {
     case PGM_TEXTURE_GST_BUFFER:
@@ -387,16 +433,12 @@ free_buffer (PgmTexture *texture)
     }
 
   texture->storage = PGM_TEXTURE_CLEAN;
-
-  PGM_TEXTURE_UNLOCK (texture);
 }
 
 /* Constructor */
 static void
 init_texture (PgmTexture *texture)
 {
-  texture->mutex = g_mutex_new ();
-
   /* Content */
   texture->storage = PGM_TEXTURE_CLEAN;
   texture->data.gstbuffer = NULL;
@@ -406,10 +448,18 @@ init_texture (PgmTexture *texture)
   texture->height_pot = -1;
   texture->stride = -1;
   texture->size = -1;
+  texture->norm_width = -1.0f;
+  texture->norm_height = -1.0f;
+  texture->inv_norm_width = -1.0f;
+  texture->inv_norm_height = -1.0f;
 
   /* Parameters */
+  texture->matrix = pgm_mat4x4_new_identity ();
   texture->filter = PGM_GL_LINEAR;
-  texture->wrap = PGM_GL_CLAMP;
+  texture->wrap = PGM_GL_CLAMP_TO_EDGE;
+
+  /* State flags */
+  texture->flags = PGM_TEXTURE_IDENTITY_MATRIX;
 
   /* Indentifiants */
   texture->count = 0;
@@ -425,7 +475,8 @@ dispose_texture (PgmTexture *texture)
   if (texture->id)
     pgm_texture_clean (texture);
 
-  g_mutex_free (texture->mutex);
+  pgm_mat4x4_free (texture->matrix);
+  texture->matrix = NULL;
 }
 
 /* Global pointers initialization */
@@ -515,8 +566,6 @@ pgm_texture_set_buffer (PgmTexture *texture,
 {
   free_buffer (texture);
 
-  PGM_TEXTURE_LOCK (texture);
-
   texture->storage = PGM_TEXTURE_BUFFER;
   texture->data.buffer = buffer;
   texture->width = width;
@@ -526,8 +575,7 @@ pgm_texture_set_buffer (PgmTexture *texture,
   texture->width_pot = get_upper_power_of_two (width);
   texture->height_pot = get_upper_power_of_two (height);
   texture->csp = csp;
-
-  PGM_TEXTURE_UNLOCK (texture);
+  update_normalized_size (texture);
 }
 
 void
@@ -536,8 +584,6 @@ pgm_texture_set_pixbuf (PgmTexture *texture,
 {
   free_buffer (texture);
 
-  PGM_TEXTURE_LOCK (texture);
-
   texture->storage = PGM_TEXTURE_PIXBUF;
   texture->data.pixbuf = gdk_pixbuf_ref (pixbuf);
   texture->width = gdk_pixbuf_get_width (pixbuf);
@@ -550,8 +596,7 @@ pgm_texture_set_pixbuf (PgmTexture *texture,
     texture->csp = PGM_IMAGE_RGBA;
   else
     texture->csp = PGM_IMAGE_RGB;
-
-  PGM_TEXTURE_UNLOCK (texture);
+  update_normalized_size (texture);
 }
 
 void
@@ -564,8 +609,6 @@ pgm_texture_set_gst_buffer (PgmTexture *texture,
 {
   free_buffer (texture);
 
-  PGM_TEXTURE_LOCK (texture);
-
   texture->storage = PGM_TEXTURE_GST_BUFFER;
   texture->data.gstbuffer = gst_buffer_ref (gstbuffer);
   texture->size = GST_BUFFER_SIZE (gstbuffer);
@@ -575,22 +618,17 @@ pgm_texture_set_gst_buffer (PgmTexture *texture,
   texture->height_pot = get_upper_power_of_two (height);
   texture->stride = stride;
   texture->csp = csp;
-
-  PGM_TEXTURE_UNLOCK (texture);
+  update_normalized_size (texture);
 }
 
 void
 pgm_texture_update_gst_buffer (PgmTexture *texture,
                                GstBuffer *gstbuffer)
 {
-  PGM_TEXTURE_LOCK (texture);
-
   if (texture->data.gstbuffer)
     gst_buffer_unref (texture->data.gstbuffer);
   texture->data.gstbuffer = gst_buffer_ref (gstbuffer);
   texture->storage = PGM_TEXTURE_GST_BUFFER;
-
-  PGM_TEXTURE_UNLOCK (texture);
 }
 
 void
@@ -598,17 +636,19 @@ pgm_texture_bind (PgmTexture *texture)
 {
   g_return_if_fail (texture != NULL);
 
-  PGM_TEXTURE_LOCK (texture);
-
   if (!texture->id)
-    {
-      PGM_TEXTURE_UNLOCK (texture);
-      return;
-    }
+    return;
 
   bind_func[texture->csp] (texture);
 
-  PGM_TEXTURE_UNLOCK (texture);
+  /* Push the current texture matrix, and load our matrix */
+  if (!(texture->flags & PGM_TEXTURE_IDENTITY_MATRIX))
+    {
+      gl->matrix_mode (PGM_GL_TEXTURE);
+      gl->push_matrix ();
+      gl->load_matrix_f (texture->matrix->m);
+      gl->matrix_mode (PGM_GL_MODELVIEW);
+    }
 }
 
 void
@@ -616,17 +656,18 @@ pgm_texture_unbind (PgmTexture *texture)
 {
   g_return_if_fail (texture != NULL);
 
-  PGM_TEXTURE_LOCK (texture);
-
   if (!texture->id)
-    {
-      PGM_TEXTURE_UNLOCK (texture);
-      return;
-    }
+    return;
 
   unbind_func[texture->csp] (texture);
 
-  PGM_TEXTURE_UNLOCK (texture);
+  /* Pop our texture matrix from the stack */
+  if (!(texture->flags & PGM_TEXTURE_IDENTITY_MATRIX))
+    {
+      gl->matrix_mode (PGM_GL_TEXTURE);
+      gl->pop_matrix ();
+      gl->matrix_mode (PGM_GL_MODELVIEW);
+    }
 }
 
 void
@@ -645,9 +686,7 @@ pgm_texture_generate (PgmTexture *texture)
   texture->id = g_slice_alloc0 (sizeof (guint) * texture->count);
   gl->gen_textures (texture->count, texture->id);
 
-  PGM_TEXTURE_LOCK (texture);
   create_func[texture->csp] (texture);
-  PGM_TEXTURE_UNLOCK (texture);
 }
 
 void
@@ -667,13 +706,8 @@ pgm_texture_upload (PgmTexture *texture)
 {
   void *buffer = NULL;
 
-  PGM_TEXTURE_LOCK (texture);
-
   if (!texture->id)
-    {
-      PGM_TEXTURE_UNLOCK (texture);
-      return;
-    }
+    return;
 
   switch (texture->storage)
     {
@@ -694,23 +728,44 @@ pgm_texture_upload (PgmTexture *texture)
     default:
       break;
     }
-
-  PGM_TEXTURE_UNLOCK (texture);
 }
 
 void
 pgm_texture_update (PgmTexture *texture)
 {
-  PGM_TEXTURE_LOCK (texture);
-
   if (!texture->id)
-    {
-      PGM_TEXTURE_UNLOCK (texture);
-      return;
-    }
+    return;
 
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
   set_texture_parameters (texture);
+}
 
-  PGM_TEXTURE_UNLOCK (texture);
+void
+pgm_texture_set_matrix (PgmTexture *texture,
+                        PgmMat4x4 *matrix)
+{
+  pgm_mat4x4_set_from_mat4x4 (texture->matrix, matrix);
+
+  /* Check identity so that we can avoid pushing our texture matrix */
+  if (!pgm_mat4x4_is_identity (matrix))
+    {
+      PgmMat4x4 *scale, *transform, *transpose;
+
+      /* We need to scale, apply the transformation and unscale to hide
+       * the use of power-of-two textures. */
+      scale = pgm_mat4x4_new_scale_from_scalars (texture->norm_width,
+                                                 texture->norm_height, 1.0f);
+      transform = pgm_mat4x4_multiply_mat4x4 (scale, texture->matrix);
+      pgm_mat4x4_scale_from_scalars (transform, texture->inv_norm_width,
+                                     texture->inv_norm_height, 1.0f);
+      transpose = pgm_mat4x4_transpose (transform);
+      pgm_mat4x4_set_from_mat4x4 (texture->matrix, transpose);
+      pgm_mat4x4_free (scale);
+      pgm_mat4x4_free (transform);
+      pgm_mat4x4_free (transpose);
+
+      texture->flags &= ~PGM_TEXTURE_IDENTITY_MATRIX;
+    }
+  else
+    texture->flags |= PGM_TEXTURE_IDENTITY_MATRIX;
 }
